###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###


from math import degrees

import pytest
from diffcalc.hkl.constraints import Constraints, _con_type, _Constraint
from diffcalc.util import DiffcalcException

from tests.test_tools import eq_
from tests.tools import assert_dict_almost_equal


def joined(d1, d2):
    d1.update(d2)
    return d1


@pytest.fixture
def cm():
    return Constraints()


def test_init(cm):
    eq_(cm.asdict, dict())
    eq_(cm._detector, dict())
    eq_(cm._reference, dict())
    eq_(cm._sample, dict())


def test_dict_init():
    cm = Constraints({"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.asdict, {"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.astuple, (("nu", 0), "a_eq_b", ("mu", 0)))
    assert cm.nu == 0
    assert cm.mu == 0
    assert cm.a_eq_b is True
    assert cm.naz is None

    cm = Constraints({"nu": 0, "mu": 0, "a_eq_b": False})
    eq_(cm.asdict, {"nu": 0, "mu": 0})
    eq_(cm.astuple, (("nu", 0), ("mu", 0)))
    eq_(cm.nu, 0)
    eq_(cm.mu, 0)
    eq_(cm.a_eq_b, None)
    eq_(cm.naz, None)

    cm.asdict = {"nu": 0, "mu": 0, "a_eq_b": None}
    eq_(cm.asdict, {"nu": 0, "mu": 0})
    eq_(cm.astuple, (("nu", 0), ("mu", 0)))
    eq_(cm.nu, 0)
    eq_(cm.mu, 0)
    eq_(cm.a_eq_b, None)
    eq_(cm.naz, None)

    cm.asdict = {"nu": 3, "mu": 4, "a_eq_b": True}
    assert_dict_almost_equal(cm.asdict, {"nu": 3, "mu": 4, "a_eq_b": True})
    eq_(cm.astuple, (("nu", pytest.approx(3)), "a_eq_b", ("mu", pytest.approx(4))))
    assert cm.nu == pytest.approx(3)
    assert cm.mu == pytest.approx(4)
    eq_(cm.a_eq_b, True)
    eq_(cm.naz, None)

    cm = Constraints((("nu", 0), ("mu", 0), "a_eq_b"))
    eq_(cm.asdict, {"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.astuple, (("nu", 0), "a_eq_b", ("mu", 0)))

    cm = Constraints([("nu", 0), ("mu", 0), "a_eq_b"])
    eq_(cm.asdict, {"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.astuple, (("nu", 0), "a_eq_b", ("mu", 0)))

    cm = Constraints({("nu", 0), ("mu", 0), "a_eq_b"})
    eq_(cm.asdict, {"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.astuple, (("nu", 0), "a_eq_b", ("mu", 0)))


def test_set_init():
    cm = Constraints({("nu", 0), ("mu", 0), "a_eq_b"})
    eq_(cm.asdict, {"nu": 0, "mu": 0, "a_eq_b": True})
    eq_(cm.astuple, (("nu", 0), "a_eq_b", ("mu", 0)))
    eq_(cm.nu, 0)
    eq_(cm.mu, 0)
    eq_(cm.a_eq_b, True)
    eq_(cm.naz, None)

    cm = Constraints({("nu", 0), ("mu", 0)})
    eq_(cm.asdict, {"nu": 0, "mu": 0})
    eq_(cm.astuple, (("nu", 0), ("mu", 0)))
    eq_(cm.nu, 0)
    eq_(cm.mu, 0)
    eq_(cm.a_eq_b, None)
    eq_(cm.naz, None)

    cm.astuple = (("nu", 3), "a_eq_b", ("mu", 4))
    assert_dict_almost_equal(cm.asdict, {"nu": 3, "mu": 4, "a_eq_b": True})

    eq_(cm.astuple, (("nu", pytest.approx(3)), "a_eq_b", ("mu", pytest.approx(4))))
    assert cm.nu == pytest.approx(3)
    assert cm.mu == pytest.approx(4)
    eq_(cm.a_eq_b, True)
    eq_(cm.naz, None)


@pytest.mark.parametrize(
    ("con_dict", "con_set", "con_list", "con_tuple"),
    [
        (
            {"nu": 3.0, "mu": 4.0, "a_eq_b": True},
            (("nu", 3.0), "a_eq_b", ("mu", 4)),
            {("nu", 3.0), ("mu", 4), "a_eq_b"},
            [("nu", 3.0), "a_eq_b", ("mu", 4)],
        ),
    ],
)
def test_all_init(con_dict, con_tuple, con_set, con_list):
    for el in (con_dict, con_tuple, con_set, con_list):
        cm = Constraints(el)
        assert_dict_almost_equal(cm.asdict, con_dict)
        # eq_(cm.astuple, con_tuple)


@pytest.mark.parametrize(
    ("con_dict"),
    [
        {"nu": 0.1, "mu": 0.2, "a_eq_b": True},
        {"nu": -0.1, "mu": 0.2, "a_eq_b": True},
    ],
)
def test_asdegrees_asradians(con_dict):
    cm = Constraints(con_dict)
    cm_deg = Constraints.asdegrees(cm)
    for k, v in cm_deg.asdict.items():
        assert cm.asdict[k] == pytest.approx(v)
    cm_rad = Constraints.asradians(cm)
    for k, v in cm_rad.asdict.items():
        if isinstance(v, bool):
            assert cm.asdict[k] == pytest.approx(v)
        else:
            assert cm.asdict[k] == pytest.approx(degrees(v))


def test_str_constraint(cm):
    print(str(cm))
    eq_(
        str(cm),
        "\n".join(
            [
                "    DET             REF             SAMP",
                "    -----------     -----------     -----------",
                "    delta           a_eq_b          mu",
                "    nu              alpha           eta",
                "    qaz             beta            chi",
                "    naz             psi             phi",
                "                    bin_eq_bout     bisect",
                "                    betain          omega",
                "                    betaout",
                "",
                "!   3 more constraints required",
                "",
            ],
        ),
    )


def test_build_display_table(cm):
    cm.qaz = 1.234
    cm.alpha = 1.0
    cm.eta = 99.0
    print("\n".join(cm._build_display_table_lines()))
    eq_(
        cm._build_display_table_lines(),
        [
            "    DET             REF             SAMP",
            "    -----------     -----------     -----------",
            "    delta           a_eq_b          mu",
            "    nu          --> alpha       --> eta",
            "--> qaz             beta            chi",
            "    naz             psi             phi",
            "                    bin_eq_bout     bisect",
            "                    betain          omega",
            "                    betaout",
        ],
    )


def test_unconstrain_okay(cm):
    eq_(cm.asdict, dict())
    cm.delta = 1.0
    cm.mu = 2
    eq_(cm._constrained, (cm._delta, cm._mu))
    eq_(cm.asdict, {"delta": 1.0, "mu": 2})
    del cm.delta
    cm.mu = None
    eq_(cm.asdict, dict())
    assert cm.delta is None


@pytest.mark.parametrize(
    "con_name",
    [
        "delta",
        "nu",
        "naz",
        "qaz",
        "a_eq_b",
        "alpha",
        "beta",
        "psi",
        "bin_eq_bout",
        "betain",
        "betaout",
        "mu",
        "eta",
        "chi",
        "phi",
        "bisect",
        "omega",
    ],
)
def test_every_constraint(con_name, cm):
    assert getattr(cm, con_name) is None
    con = getattr(cm, "_" + con_name)
    assert type(con) is _Constraint
    if con._type is _con_type.VALUE:
        setattr(cm, con_name, 1.0)
        assert getattr(cm, con_name) == 1.0
        eq_(cm.asdict, {con_name: 1.0})
    elif con._type is _con_type.VOID:
        setattr(cm, con_name, True)
        assert getattr(cm, con_name) is True
        eq_(cm.asdict, {con_name: True})
    else:
        raise TypeError("Invalid constraint type setting.")
    delattr(cm, con_name)
    assert getattr(cm, con_name) is None
    eq_(cm.asdict, {})
    eq_(cm.asdict, dict())


def test_clear_constraints(cm):
    cm.delta = 1
    cm.mu = 2
    cm.clear()
    eq_(cm.asdict, dict())
    assert cm.delta is None
    assert cm.mu is None


def test_constrain_det(cm, pre={}):
    assert_dict_almost_equal(cm.asdict, pre)
    cm.delta = 1
    assert_dict_almost_equal(cm.asdict, joined({"delta": 1}, pre))
    cm.naz = 1
    assert_dict_almost_equal(cm.asdict, joined({"naz": 1}, pre))
    cm.delta = 1
    assert_dict_almost_equal(cm.asdict, joined({"delta": 1}, pre))


def test_constrain_det_one_preexisting_ref(cm):
    cm.alpha = 2.0
    test_constrain_det(cm, {"alpha": 2.0})


def test_constrain_det_one_preexisting_samp(cm):
    cm.phi = 3.0
    test_constrain_det(cm, {"phi": 3.0})


def test_constrain_det_one_preexisting_samp_and_ref(cm):
    cm.alpha = 2.1
    cm.phi = 3.2
    test_constrain_det(cm, {"alpha": 2.1, "phi": 3.2})


def test_constrain_det_two_preexisting_samp(cm):
    cm.chi = 4.3
    cm.phi = 5.6
    test_constrain_det(cm, {"chi": 4.3, "phi": 5.6})


def test_constrain_det_three_preexisting_other(cm):
    cm.alpha = 1
    cm.phi = 2
    cm.chi = 3
    try:
        cm.delta = 4
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            (
                "Cannot set delta constraint. First un-constrain one of the"
                "\nangles alpha, chi, phi."
            ),
        )


def test_constrain_det_three_preexisting_samp(cm):
    cm.phi = 1
    cm.chi = 2
    cm.eta = 3
    try:
        cm.delta = 4
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set delta constraint. First un-constrain one of the"
            "\nangles chi, eta, phi.",
        )


def test_constrain_ref(cm, pre={}):
    assert_dict_almost_equal(cm.asdict, pre)
    cm.alpha = 1.0
    assert_dict_almost_equal(cm.asdict, joined({"alpha": 1.0}, pre))
    cm.alpha = 1.0
    assert_dict_almost_equal(cm.asdict, joined({"alpha": 1.0}, pre))
    cm.beta = 1.0
    assert_dict_almost_equal(cm.asdict, joined({"beta": 1.0}, pre))


def test_constrain_ref_one_preexisting_det(cm):
    cm.delta = 2
    test_constrain_ref(cm, {"delta": 2})


def test_constrain_ref_one_preexisting_samp(cm):
    cm.phi = 3
    test_constrain_ref(cm, {"phi": 3})


def test_constrain_ref_one_preexisting_samp_and_det(cm):
    cm.delta = 1
    cm.phi = 2
    test_constrain_ref(cm, {"delta": 1, "phi": 2})


def test_constrain_ref_two_preexisting_samp(cm):
    cm.chi = 1
    cm.phi = 2
    test_constrain_ref(cm, {"chi": 1, "phi": 2})


def test_constrain_ref_three_preexisting_other(cm):
    cm.delta = 1
    cm.phi = 2
    cm.chi = 3
    try:
        cm.alpha = 1
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set alpha constraint. First un-constrain one of the"
            "\nangles chi, delta, phi.",
        )


def test_constrain_ref_three_preexisting_samp(cm):
    cm.phi = 1
    cm.chi = 2
    cm.eta = 3
    try:
        cm.delta = 1
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set delta constraint. First un-constrain one of the"
            "\nangles chi, eta, phi.",
        )


def test_constrain_samp_when_one_free(cm, pre={}):
    assert_dict_almost_equal(cm.asdict, pre)
    cm.phi = 1.0
    assert_dict_almost_equal(cm.asdict, joined({"phi": 1.0}, pre))
    cm.phi = 1.0
    assert_dict_almost_equal(cm.asdict, joined({"phi": 1.0}, pre))


def test_constrain_samp_one_preexisting_samp(cm):
    cm.chi = 2.0
    test_constrain_samp_when_one_free(cm, {"chi": 2.0})


def test_constrain_samp_two_preexisting_samp(cm):
    cm.chi = 1
    cm.eta = 2
    test_constrain_samp_when_one_free(cm, {"chi": 1, "eta": 2})


def test_constrain_samp_two_preexisting_other(cm):
    cm.delta = 1
    cm.alpha = 2
    test_constrain_samp_when_one_free(cm, {"delta": 1, "alpha": 2})


def test_constrain_samp_two_preexisting_one_det(cm):
    cm.delta = 1
    cm.eta = 1
    test_constrain_samp_when_one_free(cm, {"delta": 1, "eta": 1})


def test_constrain_samp_two_preexisting_one_ref(cm):
    cm.alpha = 3
    cm.eta = 2
    test_constrain_samp_when_one_free(cm, {"alpha": 3, "eta": 2})


def test_constrain_samp_three_preexisting_only_one_samp(cm):
    cm.delta = 3
    cm.alpha = 4
    cm.eta = 5
    cm.phi = 1
    assert_dict_almost_equal(cm.asdict, {"delta": 3, "alpha": 4, "phi": 1})
    cm.phi = 2
    assert_dict_almost_equal(cm.asdict, {"delta": 3, "alpha": 4, "phi": 2})


def test_constrain_samp_three_preexisting_two_samp_one_det(cm):
    cm.delta = 1
    cm.eta = 2
    cm.chi = 3
    try:
        cm.phi = 4
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set phi constraint. First un-constrain one of the"
            "\nangles chi, delta, eta.",
        )


def test_constrain_samp_three_preexisting_two_samp_one_ref(cm):
    cm.alpha = 2
    cm.eta = 3
    cm.chi = 4
    try:
        cm.phi = 4
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set phi constraint. First un-constrain one of the"
            "\nangles alpha, chi, eta.",
        )


def test_constrain_samp_three_preexisting_samp(cm):
    cm.mu = 1
    cm.eta = 2
    cm.chi = 3
    try:
        cm.phi = 4
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            "Cannot set phi constraint. First un-constrain one of the"
            "\nangles chi, eta, mu.",
        )


def test_report_constraints_none(cm):
    eq_(cm._report_constraints_lines(), ["!   3 more constraints required"])


def test_report_constraints_one_with_value(cm):
    cm.nu = 9.12343
    eq_(
        cm._report_constraints_lines(),
        ["!   2 more constraints required", "    nu   : 9.1234"],
    )


def test_report_constraints_one_with_novalue(cm):
    cm.nu = None
    eq_(
        cm._report_constraints_lines(),
        ["!   3 more constraints required"],
    )


def test_report_constraints_one_with_valueless(cm):
    cm.a_eq_b = True
    eq_(
        set(cm._report_constraints_lines()),
        {"!   2 more constraints required", "    a_eq_b"},
    )


def test_report_constraints_one_with_two(cm):
    cm.naz = 9.12343
    cm.a_eq_b = True
    eq_(
        set(cm._report_constraints_lines()),
        {"!   1 more constraint required", "    naz  : 9.1234", "    a_eq_b"},
    )


def test_report_constraints_one_with_three(cm):
    cm.naz = 9.12343
    cm.a_eq_b = True
    cm.mu = 9.12343

    eq_(
        set(cm._report_constraints_lines()),
        {"    naz  : 9.1234", "    a_eq_b", "    mu   : 9.1234"},
    )


def _constrain(self, *args):
    for con in args:
        cm.constrain(con)


@pytest.mark.xfail(raises=ValueError)
def test_is_implemented_invalid(cm):
    cm.naz = 1
    cm.is_current_mode_implemented()


# 1 samp


def test_is_implemented_1_samp_naz(cm):
    cm.naz = 1
    cm.alpha = 2
    cm.mu = 3
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_1_samp_det(cm):
    cm.qaz = 1
    cm.alpha = 2
    cm.mu = 3
    eq_(cm.is_current_mode_implemented(), True)


# 2 samp + ref


def test_is_implemented_2_samp_ref_mu_chi(cm):
    cm.beta = 1
    cm.mu = 2
    cm.chi = 3
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_mu90_chi0(cm):
    cm.beta = 0
    cm.mu = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_mu_eta(cm):
    cm.beta = 0
    cm.mu = 0
    cm.eta = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_mu_phi(cm):
    cm.beta = 0
    cm.mu = 0
    cm.phi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_eta_chi(cm):
    cm.beta = 0
    cm.eta = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_eta_phi(cm):
    cm.beta = 0
    cm.eta = 0
    cm.phi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_ref_chi_phi(cm):
    cm.beta = 0
    cm.phi = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


# 2 samp + det


def test_is_implemented_2_samp_det_mu_chi(cm):
    cm.qaz = 0
    cm.mu = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_mu_eta(cm):
    cm.qaz = 0
    cm.mu = 0
    cm.eta = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_mu_phi(cm):
    cm.qaz = 0
    cm.mu = 0
    cm.phi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_eta_chi(cm):
    cm.qaz = 0
    cm.eta = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_eta_phi(cm):
    cm.qaz = 0
    cm.eta = 0
    cm.phi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_chi_phi(cm):
    cm.qaz = 0
    cm.phi = 0
    cm.chi = 90
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_bisect_mu(cm):
    cm.qaz = 0
    cm.mu = 0
    cm.bisect = True
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_bisect_eta(cm):
    cm.qaz = 0
    cm.eta = 0
    cm.bisect = True
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_2_samp_det_bisect_omega(cm):
    cm.qaz = 0
    cm.omega = 0
    cm.bisect = True
    eq_(cm.is_current_mode_implemented(), True)


# 3 samp


def test_is_implemented_3_samp_no_mu(cm):
    cm.eta = 0
    cm.chi = 0
    cm.phi = 1
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_3_samp_no_eta(cm):
    cm.mu = 0
    cm.chi = 0
    cm.phi = 1
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_3_samp_no_chi(cm):
    cm.eta = 0
    cm.chi = 0
    cm.phi = 1
    eq_(cm.is_current_mode_implemented(), True)


def test_is_implemented_3_samp_no_phi(cm):
    cm.eta = 0
    cm.mu = 0
    cm.chi = 1
    eq_(cm.is_current_mode_implemented(), True)


def test_set_fails(cm):
    try:
        cm.delta = True
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            f"Constraint delta requires numerical value. Found {bool} instead.",
        )

    try:
        cm.a_eq_b = 1
        assert False
    except DiffcalcException as e:
        eq_(
            e.args[0],
            f"Constraint a_eq_b requires boolean value. Found {int} instead.",
        )



"""UB matrix calculations.

.. currentmodule:: diffcalc.ub

Routines for defining crystal lattice and UB matrix calculation from reference
reflections and crystal reference orientations.

.. autosummary::
   :toctree: generated/

   calc - UB matrix calculation.
   crystal - Crystal lattice data object.
   fitting - Routines for refining crystal lattice parameters and U matrix.
   reference - Reference reflection and crystal orientations data object.

"""



"""Diffractometer position calculations.

.. currentmodule:: diffcalc.hkl

Routines for calculating miller indices for a given diffractometer position and
all diffractometer positions matching miller indices. This implementation
follows calculations published in [1]_.

.. autosummary::
   :toctree: generated/

   calc - Calculation of miller indices an diffractometer position.
   constraints - Diffractometer constraint management object.
   geometry - Diffractometer position object and rotation matrices.


References
----------
.. [1] H. You. "Angle calculations for a '4S+2D' six-circle diffractometer"
       J. Appl. Cryst. (1999). 32, 614-623.
"""



"""Example of calculations for (2+2) surface diffractometer.

Demonstrate creating UB calculation and setting diffractometer constraints
to calculate diffractometer position for different miller indices with
different constraints, e.g. scattering plane and reference vector orientations.
"""

from itertools import product
from pprint import pprint

import numpy as np
from diffcalc.hkl.calc import HklCalculation
from diffcalc.hkl.constraints import Constraints
from diffcalc.hkl.geometry import Position
from diffcalc.ub.calc import UBCalculation


def position_in_range(pos: Position) -> bool:
    """Check diffractometer position is in acceptale range.

    mu in (0, 90) deg
    nu in (0, 90) deg
    delta in (0, 90) deg
    phi in (-90, 90) deg

    Parameters
    ----------
    pos: Position
        Diffractometer postion to be checked.

    Returns
    -------
    bool
        True, if position is in the acceptable range.
    """
    return all(
        (
            0 < pos.mu < 90,
            0 < pos.nu < 90,
            0 < pos.delta < 90,
            -90 < pos.phi < 90,
        )
    )


def demo_hkl_positions():
    """Demonstrate calculations of miller indices and diffractometer positions."""
    cons.betain = 3.0
    for h, k, l in ((0, 0, 1), (1, 0, 1), (1, 0, 2)):
        all_pos = hklcalc.get_position(h, k, l, wavelength)
        print(f"\n{'hkl':<8s}: [{h:1.0f} {k:1.0f} {l:1.0f}]")
        for pos_001, virtual_angles in all_pos:
            if position_in_range(pos_001):
                for angle, val in pos_001.asdict.items():
                    print(f"{angle:<8s}:{val:>8.2f}")
                print("-" * 18)
                for angle, val in virtual_angles.items():
                    print(f"{angle:<8s}:{val:>8.2f}")

    pos1 = Position(3.00, 7.90, 14.79, 0.0, 0.0, 8.30)
    hkl1 = hklcalc.get_hkl(pos1, wavelength)
    print("\nPosition -> hkl")
    for angle, val in pos1.asdict.items():
        print(f"{angle:<8s}:{val:>8.2f}")
    print("-" * 18)
    print(f"\n{'hkl':<8s}: [{hkl1[0]:1.1f} {hkl1[1]:1.1f} {hkl1[2]:1.1f}]")


def demo_scan_betain(h: float, k: float, l: float) -> None:
    """Scan constrained betain incident angle at [h k l] reflection.

    Parameters
    ----------
    h: float
        Miller index
    k: float
        Miller index
    l: float
        Miller index
    """
    print(f"\n\nScanning betain incident angle at [{h} {k} {l}] reflection.\n")
    print(f"{'betain':<8s}{'mu':>12s}{'delta':>12s}{'gamma':>12s}{'phi':>12s}")
    print("-" * 56)
    for betain in np.arange(3.0, 4.1, 0.1):
        cons.betain = betain
        for pos, virtual_angles in hklcalc.get_position(h, k, l, wavelength):
            if position_in_range(pos):
                print(
                    f"{virtual_angles['betain']:<8.2f}"
                    f"{pos.mu:>12.3f}"
                    f"{pos.delta:>12.3f}"
                    f"{pos.nu:>12.3f}"
                    f"{pos.phi:>12.3f}"
                )


def demo_energy_scan(h: float, k: float, l: float) -> None:
    """Scan energy using either fixed incident angle or specular position at [h k l] reflection.

    Parameters
    ----------
    h: float
        Miller index
    k: float
        Miller index
    l: float
        Miller index
    """
    print(
        "\n\nScanning energy in 16-18 keV range at betain = 3.0 and betain=betaout constraints "
        f"at [{h} {k} {l}] reflection.\n"
    )
    for con_name, con_value in (("betain", 3.0), ("bin_eq_bout", True)):
        setattr(cons, con_name, con_value)
        print(f"\n\n{cons}")
        print(
            f"{'energy':<8s}{'mu':>12s}{'delta':>12s}{'gamma':>12s}{'phi':>12s}{'betain':>12s}{'betaout':>12s}"
        )
        print("-" * 80)
        for en in np.arange(16, 18.1, 0.2):
            pos, virtual_angles = next(
                iter(hklcalc.get_position(h, k, l, 12.39842 / en))
            )
            print(
                f"{en:<8.2f}"
                f"{pos.mu:12.3f}"
                f"{pos.delta:12.3f}"
                f"{pos.nu:12.3f}"
                f"{pos.phi:12.3f}"
                f"{virtual_angles['betain']:12.3f}"
                f"{virtual_angles['betaout']:12.3f}"
            )


def demo_scan_qaz(h, k, l):
    """Scan scattering plane azimuthal orientation.

    Parameters
    ----------
    h: float
        Miller index
    k: float
        Miller index
    l: float
        Miller index
    """
    print(
        f"\n\nScanning qaz, angle of the scattering plane azimuthal orientation at [{h} {k} {l}] reflection\n"
    )
    print(
        f"{'qaz':<8s}{'mu':>12s}{'delta':>12s}{'gamma':>12s}{'phi':>12s}{'theta':>12s}"
    )
    print("-" * 68)
    for qaz in np.arange(-1, 1.1, 0.1):
        cons.qaz = qaz
        pos, virtual_angles = next(iter(hklcalc.get_position(h, k, l, wavelength)))
        print(
            f"{virtual_angles['qaz']:<8.2f}"
            f"{pos.mu:12.3f}"
            f"{pos.delta:12.3f}"
            f"{pos.nu:12.3f}"
            f"{pos.phi:12.3f}"
            f"{virtual_angles['theta']:12.3f}"
        )


def demo_scan_hkl():
    """Scan miller indices."""
    print("\n\nScanning h and l indices in (1, 2) range with k = 0\n")
    print(
        f"{'hkl':<12s}{'mu':>12s}{'delta':>12s}{'gamma':>12s}{'phi':>12s}{'betain':>12s}{'betaout':>12s}{'theta':>12s}"
    )
    print("-" * 96)
    for h, l in product(np.arange(1, 2.1, 0.1), np.arange(1, 2.1, 0.1)):
        pos, virtual_angles = next(iter(hklcalc.get_position(h, 0, l, wavelength)))
        print(
            f"[{h:2.1f} 0 {l:2.1f}] "
            f"{pos.mu:12.3f}"
            f"{pos.delta:12.3f}"
            f"{pos.nu:12.3f}"
            f"{pos.phi:12.3f}"
            f"{virtual_angles['betain']:12.3f}"
            f"{virtual_angles['betaout']:12.3f}"
            f"{virtual_angles['theta']:12.3f}"
        )


if __name__ == "__main__":
    ubcalc = UBCalculation("surface")

    ubcalc.set_lattice("SiO2", 4.913, 5.405)

    ubcalc.n_phi = (0, 0, 1)

    # We define reciprocal lattice directions in laboratory frame
    # taking into account 2 deg crystal mismount around mu axis.
    start_pos = Position(-2.0, 0, 0, 0, 0, 0)
    ubcalc.add_orientation((0, 0, 1), (0, 0, 1), start_pos, "norm")
    ubcalc.add_orientation((0, 1, 0), (0, 1, 0), start_pos, "plane")
    ubcalc.calc_ub()

    print("UBCalculation object representation.\n")
    print(f"{ubcalc}")
    print("\nUB matrix defined as a (3, 3) NumPy array.\n")
    pprint(ubcalc.UB)

    # 2+2 sufrace diffractometer consistes of delta, nu, mu and phi angles.
    # Constraining unused eta and chi angles to 0.
    cons = Constraints({"eta": 0, "chi": 0})
    hklcalc = HklCalculation(ubcalc, cons)

    wavelength = 0.689

    demo_hkl_positions()

    demo_scan_betain(1, 0, 1)

    demo_energy_scan(0, 0, 1)

    # Remove betain = betaout reference constraint to avoid ambiguity
    # when setting qaz detector constraint.
    del cons.bin_eq_bout

    demo_scan_qaz(0, 0, 1)

    # Remove detector constraint before setting reference constraint
    del cons.qaz

    print("\n\nResetting crystal miscut to 0 (i.e. setting identity U matrix)\n")
    ubcalc.set_miscut(None, 0)
    print(f"{ubcalc}")

    demo_scan_betain(1, 0, 1)

    cons.bin_eq_bout = True
    demo_scan_hkl()

    del cons.bin_eq_bout
    demo_scan_qaz(1, 1, 1)



"""Module providing diffractometer position definition and rotation matrices.

Diffractometer axes and rotation matrix definitions are following conventions
described in [1]_.

References
----------
.. [1] H. You. "Angle calculations for a '4S+2D' six-circle diffractometer"
       J. Appl. Cryst. (1999). 32, 614-623.
"""
from math import degrees, radians
from typing import Dict, Tuple, Union

import numpy as np
from diffcalc.util import I, x_rotation, y_rotation, z_rotation
from numpy.linalg import inv


class Position:
    """Class representing diffractometer orientation.

    Diffractometer orientation corresponding to (4+2) geometry
    defined in H. You paper (add reference)

    Attributes
    ----------
    fields: Tuple[str, str, str, str, str, str]
        Tuple with angle names
    mu: float, default = 0.0
        mu angle value
    delta: float, default = 0.0
        delta angle value
    nu: float, default = 0.0
        nu angle value
    eta: float, default = 0.0
        eta angle value
    chi: float, default = 0.0
        chi angle value
    phi: float, default = 0.0
        phi angle value
    indegrees: bool, default = True
        If True, arguments are angles in degrees.
    """

    fields: Tuple[str, str, str, str, str, str] = (
        "mu",
        "delta",
        "nu",
        "eta",
        "chi",
        "phi",
    )

    def __init__(
        self,
        mu: float = 0.0,
        delta: float = 0.0,
        nu: float = 0.0,
        eta: float = 0.0,
        chi: float = 0.0,
        phi: float = 0.0,
        indegrees: bool = True,
    ):
        self._mu: float = radians(mu) if indegrees else mu
        self._delta: float = radians(delta) if indegrees else delta
        self._nu: float = radians(nu) if indegrees else nu
        self._eta: float = radians(eta) if indegrees else eta
        self._chi: float = radians(chi) if indegrees else chi
        self._phi: float = radians(phi) if indegrees else phi
        self.indegrees: bool = indegrees

    @classmethod
    def asdegrees(cls, pos: "Position") -> "Position":
        """Create new Position object with angles in degrees.

        Parameters
        ----------
        pos: Position
            Input Position object

        Returns
        -------
        Position
            New Position object with angles in degrees.
        """
        res = cls(**pos.asdict, indegrees=pos.indegrees)
        res.indegrees = True
        return res

    @classmethod
    def asradians(cls, pos: "Position") -> "Position":
        """Create new Position object with angles in radians.

        Parameters
        ----------
        pos: Position
            Input Position object

        Returns
        -------
        Position
            New Position object with angles in radians.
        """
        res = cls(**pos.asdict, indegrees=pos.indegrees)
        res.indegrees = False
        return res

    @property
    def mu(self) -> Union[float, None]:
        """Value of of mu angle."""
        if self.indegrees:
            return degrees(self._mu)
        else:
            return self._mu

    @mu.setter
    def mu(self, val):
        if self.indegrees:
            self._mu = radians(val)
        else:
            self._mu = val

    @mu.deleter
    def mu(self):
        self._mu = None

    @property
    def delta(self) -> Union[float, None]:
        """Value of of delta angle."""
        if self.indegrees:
            return degrees(self._delta)
        else:
            return self._delta

    @delta.setter
    def delta(self, val):
        if self.indegrees:
            self._delta = radians(val)
        else:
            self._delta = val

    @delta.deleter
    def delta(self):
        self._delta = None

    @property
    def nu(self) -> Union[float, None]:
        """Value of of nu angle."""
        if self.indegrees:
            return degrees(self._nu)
        else:
            return self._nu

    @nu.setter
    def nu(self, val):
        if self.indegrees:
            self._nu = radians(val)
        else:
            self._nu = val

    @nu.deleter
    def nu(self):
        self._nu = None

    @property
    def eta(self) -> Union[float, None]:
        """Value of of eta angle."""
        if self.indegrees:
            return degrees(self._eta)
        else:
            return self._eta

    @eta.setter
    def eta(self, val):
        if self.indegrees:
            self._eta = radians(val)
        else:
            self._eta = val

    @eta.deleter
    def eta(self):
        self._eta = None

    @property
    def chi(self) -> Union[float, None]:
        """Value of of chi angle."""
        if self.indegrees:
            return degrees(self._chi)
        else:
            return self._chi

    @chi.setter
    def chi(self, val):
        if self.indegrees:
            self._chi = radians(val)
        else:
            self._chi = val

    @chi.deleter
    def chi(self):
        self._chi = None

    @property
    def phi(self) -> Union[float, None]:
        """Value of of phi angle."""
        if self.indegrees:
            return degrees(self._phi)
        else:
            return self._phi

    @phi.setter
    def phi(self, val):
        if self.indegrees:
            self._phi = radians(val)
        else:
            self._phi = val

    @phi.deleter
    def phi(self):
        self._phi = None

    @property
    def asdict(self) -> Dict[str, float]:
        """Return dictionary of diffractometer angles.

        Returns
        -------
        Dict[str, float]
            Dictionary of axis names and angle values.
        """
        return {field: getattr(self, field) for field in self.fields}

    @property
    def astuple(self) -> Tuple[float, float, float, float, float, float]:
        """Return tuple of diffractometer angles.

        Returns
        -------
        Tuple[float, float, float, float, float, float]
            Tuple of angle values.
        """
        mu, delta, nu, eta, chi, phi = tuple(
            getattr(self, field) for field in self.fields
        )
        return mu, delta, nu, eta, chi, phi


def get_rotation_matrices(
    pos: Position,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray,]:
    """Create rotation matrices corresponding to the diffractometer axes.

    Parameters
    ----------
    pos: Position
        Position object containing set of diffractometer angles

    Returns
    -------
    Tuple[np.ndarray, np.ndarray,
        np.ndarray, np.ndarray,
        np.ndarray, np.ndarray]
        Tuple containing set of rotation matrices corresponding to
        input diffractometer angle values.
    """
    pos_in_rad = Position.asradians(pos)
    mu, delta, nu, eta, chi, phi = pos_in_rad.astuple
    return (
        rot_MU(mu),
        rot_DELTA(delta),
        rot_NU(nu),
        rot_ETA(eta),
        rot_CHI(chi),
        rot_PHI(phi),
    )


def rot_NU(nu: float) -> np.ndarray:
    """Return rotation matrix corresponding to nu axis.

    Parameters
    ----------
        nu: float
        nu axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return x_rotation(nu)


def rot_DELTA(delta: float) -> np.ndarray:
    """Return rotation matrix corresponding to delta axis.

    Parameters
    ----------
        delta: float
        delta axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return z_rotation(-delta)


def rot_MU(mu_or_alpha: float) -> np.ndarray:
    """Return rotation matrix corresponding to mu axis.

    Parameters
    ----------
        mu: float
        mu axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return x_rotation(mu_or_alpha)


def rot_ETA(eta: float) -> np.ndarray:
    """Return rotation matrix corresponding to eta axis.

    Parameters
    ----------
        eta: float
        eta axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return z_rotation(-eta)


def rot_CHI(chi: float) -> np.ndarray:
    """Return rotation matrix corresponding to chi axis.

    Parameters
    ----------
        chi: float
        chi axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return y_rotation(chi)


def rot_PHI(phi: float) -> np.ndarray:
    """Return rotation matrix corresponding to phi axis.

    Parameters
    ----------
        phi: float
        phi axis angle

    Returns
    -------
    np.ndarray
        Rotation matrix as a NumPy array.
    """
    return z_rotation(-phi)


def get_q_phi(pos: Position) -> np.ndarray:
    """Calculate scattering vector in laboratory frame.

    Calculate hkl in the phi frame in units of 2 * pi / lambda.

    Parameters
    ----------
    pos: object
        Diffractometer angles in radians.

    Returns
    -------
    matrix:
        Scattering vector coordinates corresponding to the input position.
    """
    pos_in_rad = Position.asradians(pos)
    [MU, DELTA, NU, ETA, CHI, PHI] = get_rotation_matrices(pos_in_rad)
    # Equation 12: Compute the momentum transfer vector in the lab  frame
    y = np.array([[0], [1], [0]])
    q_lab = (NU @ DELTA - I) @ y
    # Transform this into the phi frame.
    return inv(PHI) @ inv(CHI) @ inv(ETA) @ inv(MU) @ q_lab



###
# Copyright 2008-2019 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from diffcalc.hkl.geometry import Position
from diffcalc.ub.reference import OrientationList


class TestOrientationList:
    def setup_method(self):
        self.orientlist = OrientationList()
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        self.orientlist.add_orientation((1, 2, 3), (0.1, 0.2, 0.3), pos, "orient1")
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        self.orientlist.add_orientation(
            (1.1, 2.2, 3.3), (0.11, 0.12, 0.13), pos, "orient2"
        )

    def test_str(self):
        res = str(self.orientlist)
        assert (
            res
            == """         H     K     L       X     Y     Z        MU    DELTA       NU      ETA      CHI      PHI  TAG
   1  1.00  2.00  3.00   0.10  0.20  0.30    0.1000   0.2000   0.3000   0.4000   0.5000   0.6000  orient1
   2  1.10  2.20  3.30   0.11  0.12  0.13    0.1100   0.2200   0.3300   0.4400   0.5500   0.6600  orient2"""
        )

    def test_add_orientation(self):
        assert len(self.orientlist) == 2
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        self.orientlist.add_orientation((10, 20, 30), (0.1, 0.2, 0.3), pos, "orient1")

    def test_get_orientation(self):
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        answered = self.orientlist.get_orientation(1).astuple
        desired = ((1, 2, 3), (0.1, 0.2, 0.3), pos.astuple, "orient1")
        assert answered == desired
        answered = self.orientlist.get_orientation("orient1").astuple
        assert answered == desired

    def testRemoveOrientation(self):
        self.orientlist.remove_orientation(1)
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66).astuple
        answered = self.orientlist.get_orientation(1).astuple
        desired = ((1.1, 2.2, 3.3), (0.11, 0.12, 0.13), pos, "orient2")
        assert answered == desired
        self.orientlist.remove_orientation("orient2")
        assert self.orientlist.orientations == []

    def testedit_orientation(self):
        ps = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        self.orientlist.edit_orientation(1, (10, 20, 30), (1, 2, 3), ps, "new1")
        assert self.orientlist.get_orientation(1).astuple == (
            (10, 20, 30),
            (1, 2, 3),
            ps.astuple,
            "new1",
        )
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.orientlist.get_orientation(2).astuple == (
            (1.1, 2.2, 3.3),
            (0.11, 0.12, 0.13),
            pos.astuple,
            "orient2",
        )
        self.orientlist.edit_orientation(
            "orient2", (1.1, 2.2, 3.3), (1.11, 1.12, 1.13), pos, "new2"
        )
        assert self.orientlist.get_orientation("new2").astuple == (
            (1.1, 2.2, 3.3),
            (1.11, 1.12, 1.13),
            pos.astuple,
            "new2",
        )
        self.orientlist.edit_orientation(
            "new2", (1.1, 2.2, 3.3), (1.11, 1.12, 1.13), pos, "new1"
        )
        assert self.orientlist.get_orientation("new1").astuple == (
            (10, 20, 30),
            (1, 2, 3),
            ps.astuple,
            "new1",
        )

    def testSwapOrientation(self):
        self.orientlist.swap_orientations(1, 2)
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.orientlist.get_orientation(1).astuple == (
            (1.1, 2.2, 3.3),
            (0.11, 0.12, 0.13),
            pos.astuple,
            "orient2",
        )
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        assert self.orientlist.get_orientation(2).astuple == (
            (1, 2, 3),
            (0.1, 0.2, 0.3),
            pos.astuple,
            "orient1",
        )
        self.orientlist.swap_orientations("orient1", "orient2")
        pos = Position(0.11, 0.22, 0.33, 0.44, 0.55, 0.66)
        assert self.orientlist.get_orientation(2).astuple == (
            (1.1, 2.2, 3.3),
            (0.11, 0.12, 0.13),
            pos.astuple,
            "orient2",
        )
        pos = Position(0.1, 0.2, 0.3, 0.4, 0.5, 0.6)
        assert self.orientlist.get_orientation(1).astuple == (
            (1, 2, 3),
            (0.1, 0.2, 0.3),
            pos.astuple,
            "orient1",
        )



"""Fit crystal lattice and U matrix to reflection data.

A module implementing fitting routines for refining crystal lattice parameters
and U matrix using reflection data.
"""
from math import atan2, cos, pi, sin, sqrt
from typing import List, Sequence, Tuple

import numpy as np
from diffcalc.hkl.geometry import Position, get_rotation_matrices
from diffcalc.ub.crystal import Crystal
from diffcalc.ub.reference import Reflection
from diffcalc.util import DiffcalcException, I, angle_between_vectors, sign
from numpy.linalg import inv, norm
from scipy.optimize import minimize


def _get_refl_hkl(
    refl_list: List[Reflection],
) -> List[Tuple[np.ndarray, Position, float]]:
    refl_data = []
    for refl in refl_list:
        refl_data.append(
            (np.array([[refl.h], [refl.k], [refl.l]]), refl.pos, refl.energy)
        )
    return refl_data


def _func_crystal(
    vals: Sequence[float], uc_system: str, refl_data: Tuple[np.ndarray, Position, float]
) -> float:
    try:
        trial_cr = Crystal("trial", uc_system, *vals)
    except Exception:
        return 1e6

    res = 0
    for (hkl_vals, pos_vals, en) in refl_data:
        wl = 12.3984 / en
        [_, DELTA, NU, _, _, _] = get_rotation_matrices(pos_vals)
        q_pos = (NU @ DELTA - I) @ np.array([[0], [2 * pi / wl], [0]])
        q_hkl = trial_cr.B @ hkl_vals
        res += (norm(q_pos) - norm(q_hkl)) ** 2
    return res


def _func_orient(
    vals, crystal: Crystal, refl_data: Tuple[np.ndarray, Position, float]
) -> float:
    quat = _get_quat_from_u123(*vals)
    trial_u = _get_rot_matrix(*quat)
    tmp_ub = trial_u @ crystal.B

    res = 0.0
    for (hkl_vals, pos_vals, en) in refl_data:
        wl = 12.3984 / en
        [MU, DELTA, NU, ETA, CHI, PHI] = get_rotation_matrices(pos_vals)
        q_del = (NU @ DELTA - I) @ np.array([[0], [2 * pi / wl], [0]])
        q_vals = inv(PHI) @ inv(CHI) @ inv(ETA) @ inv(MU) @ q_del

        q_hkl = tmp_ub @ hkl_vals
        res += angle_between_vectors(q_hkl, q_vals)
    return res


def _get_rot_matrix(q0: float, q1: float, q2: float, q3: float) -> np.ndarray:
    rot = np.array(
        [
            [
                q0 ** 2 + q1 ** 2 - q2 ** 2 - q3 ** 2,
                2.0 * (q1 * q2 - q0 * q3),
                2.0 * (q1 * q3 + q0 * q2),
            ],
            [
                2.0 * (q1 * q2 + q0 * q3),
                q0 ** 2 - q1 ** 2 + q2 ** 2 - q3 ** 2,
                2.0 * (q2 * q3 - q0 * q1),
            ],
            [
                2.0 * (q1 * q3 - q0 * q2),
                2.0 * (q2 * q3 + q0 * q1),
                q0 ** 2 - q1 ** 2 - q2 ** 2 + q3 ** 2,
            ],
        ]
    )
    return rot


def _get_init_u123(um: np.ndarray) -> Tuple[float, float, float]:

    tr = um[0, 0] + um[1, 1] + um[2, 2]
    sgn_q1 = sign(um[2, 1] - um[1, 2])
    sgn_q2 = sign(um[0, 2] - um[2, 0])
    sgn_q3 = sign(um[1, 0] - um[0, 1])
    q0 = sqrt(1.0 + tr) / 2.0
    q1 = sgn_q1 * sqrt(1.0 + um[0, 0] - um[1, 1] - um[2, 2]) / 2.0
    q2 = sgn_q2 * sqrt(1.0 - um[0, 0] + um[1, 1] - um[2, 2]) / 2.0
    q3 = sgn_q3 * sqrt(1.0 - um[0, 0] - um[1, 1] + um[2, 2]) / 2.0
    u1 = (1.0 - um[0, 0]) / 2.0
    u2 = atan2(q0, q1) / (2.0 * pi)
    u3 = atan2(q2, q3) / (2.0 * pi)
    if u2 < 0:
        u2 += 1.0
    if u3 < 0:
        u3 += 1.0
    return u1, u2, u3


def _get_quat_from_u123(
    u1: float, u2: float, u3: float
) -> Tuple[float, float, float, float]:
    q0, q1 = sqrt(1.0 - u1) * sin(2.0 * pi * u2), sqrt(1.0 - u1) * cos(2.0 * pi * u2)
    q2, q3 = sqrt(u1) * sin(2.0 * pi * u3), sqrt(u1) * cos(2.0 * pi * u3)
    return q0, q1, q2, q3


def _get_uc_upper_limits(system: str) -> List[float]:
    max_unit = 100.0
    if system == "Triclinic":
        return [max_unit, max_unit, max_unit, 180.0, 180.0, 180.0]
    elif system == "Monoclinic":
        return [max_unit, max_unit, max_unit, 180.0]
    elif system == "Orthorhombic":
        return [
            max_unit,
            max_unit,
            max_unit,
        ]
    elif system == "Tetragonal" or system == "Hexagonal":
        return [
            max_unit,
            max_unit,
        ]
    elif system == "Rhombohedral":
        return [
            max_unit,
            180.0,
        ]
    elif system == "Cubic":
        return [
            max_unit,
        ]
    else:
        raise TypeError("Invalid crystal system parameter: %s" % str(system))


def fit_crystal(crystal: Crystal, refl_list: List[Reflection]) -> Crystal:
    """Fit crystal lattice parameters to reference reflections.

    Parameters
    ----------
    crystal: Crystal
        Object containing initial crysta lattice parameter estimates
    refl_list: List[Reflection]
        List or reference reflection objects

    Returns
    -------
    Crystal
        Object containing refined crystal lattice parameters.

    Raises
    ------
    DiffcalcException
        If crystal lattice object not initialised.
    """
    try:
        xtal_system, xtal_params = crystal.get_lattice_params()
        start = xtal_params
        lower = [
            0,
        ] * len(xtal_params)
        upper = _get_uc_upper_limits(xtal_system)
    except AttributeError:
        raise DiffcalcException(
            "UB matrix not initialised. Cannot run UB matrix fitting procedure."
        )

    refl_data = _get_refl_hkl(refl_list)
    bounds = list(zip(lower, upper))
    res = minimize(
        _func_crystal,
        start,
        args=(xtal_system, refl_data),
        method="SLSQP",
        tol=1e-10,
        options={"disp": False, "maxiter": 10000, "eps": 1e-6, "ftol": 1e-10},
        bounds=bounds,
    )
    vals = res.x

    res_cr = Crystal("trial", xtal_system, *vals)
    # res_cr._set_cell_for_system(uc_system, *vals)
    return res_cr


def fit_u_matrix(
    init_u: np.ndarray, crystal: Crystal, refl_list: List[Reflection]
) -> np.ndarray:
    """Fit crystal lattice parameters to reference reflections.

    Parameters
    ----------
    init_u: np.ndarray
        Initial U matrix as (3, 3) NumPy array
    crystal: Crystal
        Object containing initial crystal lattice parameter estimates
    refl_list: List[Reflection]
        List or reference reflection objects

    Returns
    -------
    np.ndarray
        NumPy array with refined U matrix elements.

    Raises
    ------
    DiffcalcException
        If U matrix is not initialised.
    """
    try:
        start = list(_get_init_u123(init_u))
        lower = [0, 0, 0]
        upper = [1, 1, 1]
    except TypeError:
        raise DiffcalcException(
            "UB matrix not initialised. Cannot run UB matrix fitting procedure."
        )

    ref_data = _get_refl_hkl(refl_list)
    bounds = list(zip(lower, upper))
    res = minimize(
        _func_orient,
        start,
        args=(crystal, ref_data),
        method="SLSQP",
        tol=1e-10,
        options={"disp": False, "maxiter": 10000, "eps": 1e-6, "ftol": 1e-10},
        bounds=bounds,
    )
    vals = res.x

    q0, q1, q2, q3 = _get_quat_from_u123(*vals)
    res_u = _get_rot_matrix(q0, q1, q2, q3)
    # angle = 2. * acos(q0)
    # xr = q1 / sqrt(1. - q0 * q0)
    # yr = q2 / sqrt(1. - q0 * q0)
    # zr = q3 / sqrt(1. - q0 * q0)
    # print angle * TODEG, (xr, yr, zr), res
    return res_u



"""Logging configuration."""
import getpass
import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)s:%(name)s:%(message)s",
    datefmt="%m/%d/%Y %I:%M:%S",
    filename="/tmp/diffcalc_%s.log" % getpass.getuser(),
    level=logging.DEBUG,
)



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

import math
from math import degrees
from unittest.mock import Mock

from diffcalc.hkl.calc import HklCalculation
from diffcalc.hkl.geometry import Position
from diffcalc.ub.calc import UBCalculation
from diffcalc.util import I
from numpy import array

from tests.tools import assert_almost_equal, assert_matrix_almost_equal

x = array([[1], [0], [0]])
y = array([[0], [1], [0]])
z = array([[0], [0], [1]])


def isnan(n):
    # math.isnan was introduced only in python 2.6 and is not in Jython (2.5.2)
    try:
        return math.isnan(n)
    except AttributeError:
        return n != n  # for Jython


class Test_position_to_virtual_angles:
    def setup_method(self):
        constraints = Mock()
        constraints.is_fully_constrained.return_value = True
        self.ubcalc = UBCalculation()
        self.ubcalc.set_lattice("xtal", 1)
        self.ubcalc.set_u(I)
        self.ubcalc.n_phi = (0, 0, 1)
        self.calc = HklCalculation(self.ubcalc, constraints)

    def check_angle(
        self, name, expected, mu=-99, delta=99, nu=99, eta=99, chi=99, phi=99
    ):
        """All in degrees"""
        pos = Position(mu=mu, delta=delta, nu=nu, eta=eta, chi=chi, phi=phi)
        calculated = self.calc.get_virtual_angles(pos, False)[name]
        assert_almost_equal(degrees(calculated), expected)

    # theta

    def test_theta0(self):
        self.check_angle("theta", 0, delta=0, nu=0)

    def test_theta1(self):
        self.check_angle("theta", 1, delta=2, nu=0)

    def test_theta2(self):
        self.check_angle("theta", 1, delta=0, nu=2)

    def test_theta3(self):
        self.check_angle("theta", 1, delta=-2, nu=0)

    def test_theta4(self):
        self.check_angle("theta", 1, delta=0, nu=-2)

    # qaz

    def test_qaz0_degenerate_case(self):
        self.check_angle("qaz", 0, delta=0, nu=0)

    def test_qaz1(self):
        self.check_angle("qaz", 90, delta=2, nu=0)

    def test_qaz2(self):
        self.check_angle("qaz", 90, delta=90, nu=0)

    def test_qaz3(self):
        self.check_angle(
            "qaz",
            0,
            delta=0,
            nu=1,
        )

    # Can't see one by eye
    # def test_qaz4(self):
    #    pos = Pos(delta=20*TORAD, nu=20*TORAD)#.inRadians()
    #    assert_almost_equal(
    #        self.calc._anglesToVirtualAngles(pos, None)['qaz']*TODEG, 45)

    # alpha
    def test_defaultReferenceValue(self):
        # The following tests depemd on this
        assert_matrix_almost_equal(self.calc.ubcalc.n_phi, array([[0], [0], [1]]))

    def test_alpha0(self):
        self.check_angle("alpha", 0, mu=0, eta=0, chi=0, phi=0)

    def test_alpha1(self):
        self.check_angle("alpha", 0, mu=0, eta=0, chi=0, phi=10)

    def test_alpha2(self):
        self.check_angle("alpha", 0, mu=0, eta=0, chi=0, phi=-10)

    def test_alpha3(self):
        self.check_angle("alpha", 2, mu=2, eta=0, chi=0, phi=0)

    def test_alpha4(self):
        self.check_angle("alpha", -2, mu=-2, eta=0, chi=0, phi=0)

    def test_alpha5(self):
        self.check_angle("alpha", 2, mu=0, eta=90, chi=2, phi=0)

    # beta

    def test_beta0(self):
        self.check_angle("beta", 0, delta=0, nu=0, mu=0, eta=0, chi=0, phi=0)

    def test_beta1(self):
        self.check_angle("beta", 0, delta=10, nu=0, mu=0, eta=6, chi=0, phi=5)

    def test_beta2(self):
        self.check_angle("beta", 10, delta=0, nu=10, mu=0, eta=0, chi=0, phi=0)

    def test_beta3(self):
        self.check_angle("beta", -10, delta=0, nu=-10, mu=0, eta=0, chi=0, phi=0)

    def test_beta4(self):
        self.check_angle("beta", 5, delta=0, nu=10, mu=5, eta=0, chi=0, phi=0)

    # azimuth
    def test_naz0(self):
        self.check_angle("naz", 0, mu=0, eta=0, chi=0, phi=0)

    def test_naz1(self):
        self.check_angle("naz", 0, mu=0, eta=0, chi=0, phi=10)

    def test_naz3(self):
        self.check_angle("naz", 0, mu=10, eta=0, chi=0, phi=10)

    def test_naz4(self):
        self.check_angle("naz", 2, mu=0, eta=0, chi=2, phi=0)

    def test_naz5(self):
        self.check_angle("naz", -2, mu=0, eta=0, chi=-2, phi=0)

    # tau
    def test_tau0(self):
        self.check_angle("tau", 0, mu=0, delta=0, nu=0, eta=0, chi=0, phi=0)
        # self.check_angle('tau_from_dot_product', 90, mu=0, delta=0,
        # nu=0, eta=0, chi=0, phi=0)

    def test_tau1(self):
        self.check_angle("tau", 90, mu=0, delta=20, nu=0, eta=10, chi=0, phi=0)
        # self.check_angle('tau_from_dot_product', 90, mu=0, delta=20,
        # nu=0, eta=10, chi=0, phi=0)

    def test_tau2(self):
        self.check_angle("tau", 90, mu=0, delta=20, nu=0, eta=10, chi=0, phi=3)
        # self.check_angle('tau_from_dot_product', 90, mu=0, delta=20,
        # nu=0, eta=10, chi=0, phi=3)

    def test_tau3(self):
        self.check_angle("tau", 88, mu=0, delta=20, nu=0, eta=10, chi=2, phi=0)
        # self.check_angle('tau_from_dot_product', 88, mu=0, delta=20,
        # nu=0, eta=10, chi=2, phi=0)

    def test_tau4(self):
        self.check_angle("tau", 92, mu=0, delta=20, nu=0, eta=10, chi=-2, phi=0)
        # self.check_angle('tau_from_dot_product', 92, mu=0, delta=20,
        # nu=0, eta=10, chi=-2, phi=0)

    def test_tau5(self):
        self.check_angle("tau", 10, mu=0, delta=0, nu=20, eta=0, chi=0, phi=0)
        # self.check_angle('tau_from_dot_product', 10, mu=0, delta=0,
        # nu=20, eta=0, chi=0, phi=0)

    # psi

    def test_psi0(self):
        pos = Position()
        assert isnan(self.calc.get_virtual_angles(pos)["psi"])

    def test_psi1(self):
        self.check_angle("psi", 90, mu=0, delta=11, nu=0, eta=0, chi=0, phi=0)

    def test_psi2(self):
        self.check_angle("psi", 100, mu=10, delta=0.001, nu=0, eta=0, chi=0, phi=0)

    def test_psi3(self):
        self.check_angle("psi", 80, mu=-10, delta=0.001, nu=0, eta=0, chi=0, phi=0)

    def test_psi4(self):
        self.check_angle("psi", 90, mu=0, delta=11, nu=0, eta=0, chi=0, phi=12.3)

    def test_psi5(self):
        # self.check_angle('psi', 0, mu=10, delta=.00000001,
        # nu=0, eta=0, chi=90, phi=0)
        pos = Position(mu=0, delta=0, nu=0, eta=0, chi=90, phi=0)
        assert isnan(self.calc.get_virtual_angles(pos)["psi"])

    def test_psi6(self):
        self.check_angle("psi", 90, mu=0, delta=0.001, nu=0, eta=90, chi=0, phi=0)

    def test_psi7(self):
        self.check_angle("psi", 92, mu=0, delta=0.001, nu=0, eta=90, chi=2, phi=0)

    def test_psi8(self):
        self.check_angle("psi", 88, mu=0, delta=0.001, nu=0, eta=90, chi=-2, phi=0)



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from math import degrees, radians, sqrt

import pytest
from diffcalc.hkl.calc import HklCalculation
from diffcalc.hkl.constraints import Constraints
from diffcalc.hkl.geometry import Position
from diffcalc.ub.calc import ReferenceVector, UBCalculation
from diffcalc.util import DiffcalcException
from numpy import array

from tests.diffcalc import scenarios
from tests.test_tools import eq_
from tests.tools import assert_iterable_almost_equal, mneq_


class TestUBCalculation:
    def setup_method(self):
        self._refineub_matrix = array(
            [
                [0.70711, 0.70711, 0.00000],
                [-0.70711, 0.70711, 0.00000],
                [0.00000, 0.00000, 1.00000],
            ]
        )

    def test_str(self):
        ubcalc = UBCalculation("test_str")
        ubcalc.n_phi = (0, 0, 1)
        ubcalc.surf_nphi = (0, 0, 1)
        ubcalc.set_lattice("xtal", "Cubic", 1)
        ubcalc.add_reflection((0, 0, 1), Position(0, 60, 0, 30, 0, 0), 12.4, "ref1")
        ubcalc.add_orientation(
            (0, 1, 0), (0, 1, 0), Position(1, 0, 0, 0, 2, 0), "orient1"
        )
        ubcalc.set_miscut(None, radians(2.0))

        assert (
            str(ubcalc)
            == """UBCALC

   name:      test_str

REFERNCE

   n_hkl:     -0.03490   0.00000   0.99939
   n_phi:      0.00000   0.00000   1.00000 <- set

SURFACE NORMAL

   n_hkl:     -0.03490   0.00000   0.99939
   n_phi:      0.00000   0.00000   1.00000 <- set

CRYSTAL

   name:          xtal

   a, b, c:    1.00000   1.00000   1.00000
              90.00000  90.00000  90.00000  Cubic

   B matrix:   6.28319   0.00000   0.00000
               0.00000   6.28319   0.00000
               0.00000   0.00000   6.28319

UB MATRIX

   U matrix:   0.99939   0.00000   0.03490
               0.00000   1.00000   0.00000
              -0.03490   0.00000   0.99939

   miscut:
      angle:   2.00000
       axis:   0.00000   1.00000   0.00000

   UB matrix:  6.27936   0.00000   0.21928
               0.00000   6.28319   0.00000
              -0.21928   0.00000   6.27936

REFLECTIONS

     ENERGY     H     K     L        MU    DELTA       NU      ETA      CHI      PHI  TAG
   1 12.400  0.00  0.00  1.00    0.0000  60.0000   0.0000  30.0000   0.0000   0.0000  ref1

CRYSTAL ORIENTATIONS

         H     K     L       X     Y     Z        MU    DELTA       NU      ETA      CHI      PHI  TAG
   1  0.00  1.00  0.00   0.00  1.00  0.00    1.0000   0.0000   0.0000   0.0000   2.0000   0.0000  orient1"""
        )

    def test_set_lattice(self):
        ubcalc = UBCalculation("testing_set_lattice")
        with pytest.raises(TypeError):
            ubcalc.set_lattice(1)
        with pytest.raises(TypeError):
            ubcalc.set_lattice(1, 2)
        with pytest.raises(TypeError):
            ubcalc.set_lattice("HCl")
        ubcalc.set_lattice("NaCl", 1.1)
        eq_(("NaCl", 1.1, 1.1, 1.1, 90, 90, 90), ubcalc.crystal.get_lattice())
        ubcalc.set_lattice("NaCl", 1.1, 2.2)
        eq_(("NaCl", 1.1, 1.1, 2.2, 90, 90, 90), ubcalc.crystal.get_lattice())
        ubcalc.set_lattice("NaCl", 1.1, 2.2, 3.3)
        eq_(("NaCl", 1.1, 2.2, 3.3, 90, 90, 90), ubcalc.crystal.get_lattice())
        ubcalc.set_lattice("NaCl", 1.1, 2.2, 3.3, 91)
        eq_(("NaCl", 1.1, 2.2, 3.3, 90, 91, 90), ubcalc.crystal.get_lattice())
        with pytest.raises(TypeError):
            ubcalc.set_lattice(("NaCl", 1.1, 2.2, 3.3, 91, 92))
        ubcalc.set_lattice("NaCl", 1.1, 2.2, 3.3, 91, 92, 93)
        assert_iterable_almost_equal(
            ("NaCl", 1.1, 2.2, 3.3, 91, 92, 92.99999999999999),
            ubcalc.crystal.get_lattice(),
        )

    def test_add_reflection(self):
        # start new ubcalc
        ubcalc = UBCalculation("testing_add_reflection")
        reflist = ubcalc.reflist  # for convenience

        pos1 = Position(1.1, 1.2, 1.3, 1.4, 1.5, 1.6)
        pos2 = Position(2.1, 2.2, 2.3, 2.4, 2.5, 2.6)
        pos3 = (3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 1000)
        pos4 = Position(2.1, 2.2, 2.3, 2.4, 2.5, 2.6)
        #
        energy = 1.10
        ubcalc.add_reflection((1.1, 1.2, 1.3), pos1, energy)
        result = reflist.get_reflection(1)
        eq_(result.astuple, ((1.1, 1.2, 1.3), pos1.astuple, 1.10, None))

        energy = 2.10
        ubcalc.add_reflection((2.1, 2.2, 2.3), pos2, energy, "atag")
        result = reflist.get_reflection(2)
        eq_(result.astuple, ((2.1, 2.2, 2.3), pos2.astuple, 2.10, "atag"))

        with pytest.raises(TypeError):
            ubcalc.add_reflection((3.1, 3.2, 3.3), pos3, 3.10)
        with pytest.raises(IndexError):
            reflist.get_reflection(3)

        ubcalc.add_reflection((4.1, 4.2, 4.3), pos4, 4.10, "tag2")
        result = reflist.get_reflection(3)
        eq_(result.astuple, ((4.1, 4.2, 4.3), pos4.astuple, 4.10, "tag2"))

    def test_edit_reflection(self):
        pos1 = Position(1.1, 1.2, 1.3, 1.4, 1.5, 1.6)
        pos2 = Position(2.1, 2.2, 2.3, 2.4, 2.5, 2.6)
        ubcalc = UBCalculation("testing_editref")
        ubcalc.add_reflection((1, 2, 3), pos1, 10, "tag1")
        energy = 11
        ubcalc.edit_reflection(1, (1.1, 2, 3.1), pos2, energy, "tag1")

        reflist = ubcalc.reflist  # for convenience
        result = reflist.get_reflection(1)
        eq_(result.astuple, ((1.1, 2, 3.1), pos2.astuple, 11, "tag1"))

    def test_swap_reflections(self):
        ubcalc = UBCalculation("testing_swapref")
        pos = Position(1.1, 1.2, 1.3, 1.4, 1.5, 1.6)
        ubcalc.add_reflection((1, 2, 3), pos, 10, "tag1")
        ubcalc.add_reflection((1, 2, 3), pos, 10, "tag2")
        ubcalc.add_reflection((1, 2, 3), pos, 10, "tag3")
        ubcalc.swap_reflections(1, 3)
        ubcalc.swap_reflections(1, 3)
        ubcalc.swap_reflections(3, 1)  # end flipped
        reflist = ubcalc.reflist
        tag1 = reflist.get_reflection(1).tag
        tag2 = reflist.get_reflection(2).tag
        tag3 = reflist.get_reflection(3).tag
        eq_(tag1, "tag3")
        eq_(tag2, "tag2")
        eq_(tag3, "tag1")
        ubcalc.swap_reflections(1, 2)
        tag1 = reflist.get_reflection(1).tag
        tag2 = reflist.get_reflection(2).tag
        eq_(tag1, "tag2")
        eq_(tag2, "tag3")

    def test_delref(self):
        ubcalc = UBCalculation("testing_swapref")
        pos = Position(1.1, 1.2, 1.3, 1.4, 1.5, 1.6)
        ubcalc.add_reflection((1, 2, 3), pos, 10, "tag1")
        reflist = ubcalc.reflist
        reflist.get_reflection(1)
        ubcalc.del_reflection(1)
        with pytest.raises(IndexError):
            reflist.get_reflection(1)
        with pytest.raises(IndexError):
            ubcalc.del_reflection(1)

    def test_add_orientation(self):
        ubcalc = UBCalculation("testing_add_orientation")
        orientlist = ubcalc.orientlist  # for convenience

        hkl1 = (1.1, 1.2, 1.3)
        hkl2 = (2.1, 2.2, 2.3)
        orient1 = (1.4, 1.5, 1.6)
        orient2 = (2.4, 2.5, 2.6)
        #
        ubcalc.add_orientation(hkl1, orient1)
        result = orientlist.get_orientation(1)
        trans_orient1 = array([orient1])
        eq_((result.h, result.k, result.l), hkl1)
        mneq_(array([[result.x, result.y, result.z]]), trans_orient1)
        eq_(result.tag, None)

        with pytest.raises(TypeError):
            ubcalc.add_orientation(hkl2, orient2, "atag")

        ubcalc.add_orientation(hkl2, orient2, tag="atag")
        result = orientlist.get_orientation(2)
        trans_orient2 = array([orient2])
        eq_((result.h, result.k, result.l), hkl2)
        mneq_(array([[result.x, result.y, result.z]]), trans_orient2)
        eq_(result.tag, "atag")

    def test_edit_orientation(self):
        hkl1 = (1.1, 1.2, 1.3)
        hkl2 = (1.1, 1.2, 3.1)
        orient1 = (1.4, 1.5, 1.6)
        orient2 = (2.4, 1.5, 2.6)
        ubcalc = UBCalculation("testing_editorient")
        ubcalc.add_orientation(hkl1, orient1, tag="tag1")
        ubcalc.edit_orientation(1, (1.1, 1.2, 3.1), orient2, None, "newtag")

        orientlist = ubcalc.orientlist
        result = orientlist.get_orientation(1)
        trans_orient2 = array([orient2])
        eq_((result.h, result.k, result.l), hkl2)
        mneq_(array([[result.x, result.y, result.z]]), trans_orient2)
        eq_(result.tag, "newtag")

    def test_swap_orientations(self):
        ubcalc = UBCalculation("testing_swap_orientations")
        hkl = (1.1, 1.2, 1.3)
        orient = (1.4, 1.5, 1.6)
        ubcalc.add_orientation(hkl, orient, tag="tag1")
        ubcalc.add_orientation(hkl, orient, tag="tag2")
        ubcalc.add_orientation(hkl, orient, tag="tag3")
        ubcalc.swap_orientations(1, 3)
        ubcalc.swap_orientations(1, 3)
        ubcalc.swap_orientations(3, 1)  # end flipped
        orientlist = ubcalc.orientlist
        tag1 = orientlist.get_orientation(1).tag
        tag2 = orientlist.get_orientation(2).tag
        tag3 = orientlist.get_orientation(3).tag
        eq_(tag1, "tag3")
        eq_(tag2, "tag2")
        eq_(tag3, "tag1")
        ubcalc.swap_orientations(1, 2)
        tag1 = orientlist.get_orientation(1).tag
        tag2 = orientlist.get_orientation(2).tag
        eq_(tag1, "tag2")
        eq_(tag2, "tag3")

    def test_del_orientation(self):
        ubcalc = UBCalculation("testing_del_orientation")
        hkl = (1.1, 1.2, 1.3)
        pos = (1.4, 1.5, 1.6)
        ubcalc.add_orientation(hkl, pos, tag="tag1")
        orientlist = ubcalc.orientlist
        orientlist.get_orientation(1)
        ubcalc.del_orientation(1)
        with pytest.raises(IndexError):
            orientlist.get_orientation(1)
        with pytest.raises(IndexError):
            ubcalc.del_orientation(1)

    def test_setu(self):
        # just test calling this method
        # self.ub.setu([[1,2,3],[1,2,3],[1,2,3]])
        ubcalc = UBCalculation("test_setu")
        setu = ubcalc.set_u
        with pytest.raises(TypeError):
            setu(1, 2)
        with pytest.raises(TypeError):
            setu(1)
        with pytest.raises(ValueError):
            setu("a")
        with pytest.raises(TypeError):
            setu([1, 2, 3])
        with pytest.raises(ValueError):
            setu([[1.0, 2.0, 3.0], [1.0, 2.0, 3.0], [1.0, 2.0]])
        # diffCalcException expected if no lattice set yet
        setu([[1, 2, 3], [1, 2, 3], [1, 2, 3]])  # check no exceptions only
        setu(((1, 2, 3), (1, 2, 3), (1, 2, 3)))  # check no exceptions only

    def test_setub(self):
        # just test calling this method
        ubcalc = UBCalculation("test_setub")
        setub = ubcalc.set_ub
        with pytest.raises(TypeError):
            setub(1)
        with pytest.raises(ValueError):
            setub("a")
        with pytest.raises(TypeError):
            setub([1, 2, 3])
        with pytest.raises(ValueError):
            setub([[1, 2, 3], [1, 2, 3], [1, 2]])
        setub([[1, 2, 3], [1, 2, 3], [1, 2, 3]])  # check no exceptions only
        setub(((1, 2, 3), (1, 2, 3), (1, 2, 3)))  # check no exceptions only

    def test_calcub(self):
        # not enough reflections:
        ubcalc = UBCalculation("test_calcub")
        with pytest.raises(DiffcalcException):
            ubcalc.calc_ub()

        for s in scenarios.sessions():
            ubcalc = UBCalculation("test_calcub")
            ubcalc.set_lattice(s.name, *s.lattice)
            r = s.ref1
            ubcalc.add_reflection((r.h, r.k, r.l), r.pos, r.energy, r.tag)
            r = s.ref2
            ubcalc.add_reflection((r.h, r.k, r.l), r.pos, r.energy, r.tag)
            ubcalc.calc_ub(s.ref1.tag, s.ref2.tag)
            mneq_(
                ubcalc.UB,
                array(s.umatrix) @ array(s.bmatrix),
                4,
                note="wrong UB matrix after calculating U",
            )

    def test_orientub(self):
        ubcalc = UBCalculation("test_orientub")
        # not enough orientations:
        with pytest.raises(DiffcalcException):
            ubcalc.calc_ub()

        s = scenarios.sessions()[1]
        ubcalc.set_lattice(s.name, *s.lattice)
        r1 = s.ref1
        orient1 = array([[1], [0], [0]]).T.tolist()[0]
        tag1 = "or" + r1.tag
        ubcalc.add_orientation((r1.h, r1.k, r1.l), orient1, tag=tag1)
        r2 = s.ref2
        orient2 = array([[0], [-1], [0]]).T.tolist()[0]
        tag2 = "or" + r2.tag
        ubcalc.add_orientation((r2.h, r2.k, r2.l), orient2, tag=tag2)
        ubcalc.calc_ub()
        mneq_(
            ubcalc.UB,
            array(s.umatrix) @ array(s.bmatrix),
            4,
            note="wrong UB matrix after calculating U",
        )
        ubcalc.calc_ub(tag1, tag2)
        mneq_(
            ubcalc.UB,
            array(s.umatrix) * array(s.bmatrix),
            4,
            note="wrong UB matrix after calculating U",
        )
        ubcalc.add_reflection((r1.h, r1.k, r1.l), r1.pos, r1.energy, r1.tag)
        ubcalc.calc_ub(r1.tag, tag2)
        mneq_(
            ubcalc.UB,
            array(s.umatrix) @ array(s.bmatrix),
            4,
            note="wrong UB matrix after calculating U",
        )
        ubcalc.add_reflection((r2.h, r2.k, r2.l), r2.pos, r2.energy, r2.tag)
        ubcalc.calc_ub(tag1, r2.tag)
        mneq_(
            ubcalc.UB,
            array(s.umatrix) @ array(s.bmatrix),
            4,
            note="wrong UB matrix after calculating U",
        )

    def test_refine_ub(self):
        ubcalc = UBCalculation("testing_refine_ub")
        ubcalc.set_lattice("xtal", 1, 1, 1, 90, 90, 90)
        ubcalc.set_miscut(None, 0)
        ubcalc.refine_ub(
            (1, 1, 0),
            Position(mu=0, delta=60, nu=0, eta=30, chi=0, phi=0),
            1.0,
            True,
            True,
        )
        eq_(("xtal", sqrt(2.0), sqrt(2.0), 1, 90, 90, 90), ubcalc.crystal.get_lattice())
        mneq_(
            ubcalc.U,
            self._refineub_matrix,
            4,
            note="wrong U matrix after refinement",
        )

    def test_fit_ub(self):
        for s in scenarios.sessions()[-3:]:
            ubcalc = UBCalculation("test_fit_ub_matrix")
            a, b, c, alpha, beta, gamma = s.lattice
            for r in s.reflist:
                ubcalc.add_reflection((r.h, r.k, r.l), r.pos, r.energy, r.tag)
            ubcalc.set_lattice(s.name, s.system, *s.lattice)
            ubcalc.calc_ub(s.ref1.tag, s.ref2.tag)

            init_latt = (
                1.06 * a,
                1.07 * b,
                0.94 * c,
                1.05 * alpha,
                1.06 * beta,
                0.95 * gamma,
            )
            ubcalc.set_lattice(s.name, s.system, *init_latt)
            ubcalc.set_miscut([0.2, 0.8, 0.1], 3.0, True)

            ubcalc.fit_ub([r.tag for r in s.reflist], True, True)
            assert ubcalc.crystal.system == s.system
            mneq_(
                array([ubcalc.crystal.get_lattice()[1:]]),
                array([s.lattice]),
                2,
                note="wrong lattice after fitting UB",
            )
            mneq_(
                ubcalc.U,
                array(s.umatrix),
                3,
                note="wrong U matrix after fitting UB",
            )

    def test_get_ttheta_from_hkl(self):
        ubcalc = UBCalculation("test_get_ttheta_from_hkl")
        ubcalc.set_lattice("cube", 1, 1, 1, 90, 90, 90)
        assert ubcalc.get_ttheta_from_hkl((0, 0, 1), 12.39842) == pytest.approx(
            radians(60)
        )

    def test_miscut(self):
        ubcalc = UBCalculation("testsetmiscut")
        ubcalc.reference = ReferenceVector((0, 0, 1), False)
        ubcalc.set_lattice("cube", 1, 1, 1, 90, 90, 90)
        beam_axis = array([[0], [1], [0]]).T.tolist()[0]
        beam_maxis = array([[0], [-1], [0]]).T.tolist()[0]
        ubcalc.set_miscut(beam_axis, radians(30))
        mneq_(
            ubcalc.n_hkl,
            array([[-0.5000000], [0.00000], [0.8660254]]),
        )
        ubcalc.set_miscut(beam_axis, radians(15), True)
        mneq_(
            ubcalc.n_hkl,
            array([[-0.7071068], [0.00000], [0.7071068]]),
        )
        ubcalc.set_miscut(beam_maxis, radians(45), True)
        mneq_(ubcalc.n_hkl, array([[0.0], [0.0], [1.0]]))

    @pytest.mark.parametrize(
        ("axis", "angle", "hkl"),
        [((0, 1, 0), 10, (0, 0, 1)), ((1, 0, 0), 30, (0, 1, 1))],
    )
    def test_get_miscut_from_hkl(self, axis, angle, hkl):
        ubcalc = UBCalculation("testing_calc_miscut")
        ubcalc.set_lattice("xtal", 1, 1, 1, 90, 90, 90)
        ubcalc.set_miscut(axis, radians(angle))
        hklcalc = HklCalculation(ubcalc, Constraints({"delta": 0, "psi": 0, "eta": 0}))
        pos, _ = hklcalc.get_position(*hkl, 1.0)[0]
        ubcalc.set_miscut(None, 0)
        miscut, miscut_axis = ubcalc.get_miscut_from_hkl(hkl, pos)
        assert miscut == pytest.approx(angle)
        mneq_(
            array([axis]),
            array([miscut_axis]),
            4,
            note="wrong calculation for miscut axis",
        )

    @pytest.mark.parametrize(
        ("axis", "angle"),
        [((0, 1, 0), 10), ((1, 0, 0), 30), ((0.70710678, 0.70710678, 0), 50)],
    )
    def test_get_miscut(self, axis, angle):
        ubcalc = UBCalculation("testing_calc_miscut")
        ubcalc.set_lattice("xtal", 1, 1, 1, 90, 90, 90)
        ubcalc.set_miscut(axis, radians(angle))
        test_angle, test_axis = ubcalc.get_miscut()
        assert degrees(test_angle) == pytest.approx(angle)
        mneq_(
            array([test_axis]),
            array([axis]),
            4,
            note="wrong calculation for miscut axis",
        )



"""Collection of auxiliary mathematical methods."""
from math import acos, cos, isclose, sin
from typing import Any, Sequence, Tuple

import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform.rotation import Rotation

I: np.ndarray = np.identity(3)

SMALL: float = 1e-7


def x_rotation(th: float) -> np.ndarray:
    """Rotation matrix over x axis.

    Parameters
    ----------
    th: float
        Rotation angle.

    Returns
    -------
    np.ndarray
        Rotation matrix.
    """
    return np.array(((1, 0, 0), (0, cos(th), -sin(th)), (0, sin(th), cos(th))))


def y_rotation(th: float) -> np.ndarray:
    """Rotation matrix over y axis.

    Parameters
    ----------
    th: float
        Rotation angle.

    Returns
    -------
    np.ndarray
        Rotation matrix.
    """
    return np.array(((cos(th), 0, sin(th)), (0, 1, 0), (-sin(th), 0, cos(th))))


def z_rotation(th: float) -> np.ndarray:
    """Rotation matrix over z axis.

    Parameters
    ----------
    th: float
        Rotation angle.

    Returns
    -------
    np.ndarray
        Rotation matrix.
    """
    return np.array(((cos(th), -sin(th), 0), (sin(th), cos(th), 0), (0, 0, 1)))


def xyz_rotation(axis: Tuple[float, float, float], angle: float) -> np.ndarray:
    """Rotation matrix over arbitrary axis.

    Parameters
    ----------
    axis: Tuple[float, float, float]
        Rotation axis coordinates
    angle: float
        Rotation angle.

    Returns
    -------
    np.ndarray
        Rotation matrix.
    """
    rot = Rotation.from_rotvec(angle * np.array(axis) / norm(np.array(axis)))
    return rot.as_matrix()


class DiffcalcException(Exception):
    """Error caused by user misuse of diffraction calculator."""

    def __str__(self):
        """Error message."""
        lines = []
        message = super().__str__()
        for msg_line in message.split("\n"):
            lines.append("* " + msg_line)
        width = max(len(line) for line in lines)
        lines.insert(0, "\n\n" + "*" * width)
        lines.append("*" * width)
        return "\n".join(lines)


### Matrices


def cross3(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Cross product of column vectors.

    Parameters
    ----------
    x: np.ndarray
        Column vector represented as NumPy (3,1) array.
    y: np.ndarray
        Column vector represented as NumPy (3,1) array.

    Returns
    -------
    np.ndarray
        Cross product column vector as as NumPy (3,1) array.
    """
    v = np.cross(x.T[0], y.T[0])
    return np.array([v]).T


def dot3(x: np.ndarray, y: np.ndarray) -> float:
    """Dot product of column vectors.

    Parameters
    ----------
    x: np.ndarray
        Column vector represented as NumPy (3,1) array.
    y: np.ndarray
        Column vector represented as NumPy (3,1) array.

    Returns
    -------
    float
        Dot product of column vectors.
    """
    return float(np.dot(x.T[0], y.T[0]))


def angle_between_vectors(x: np.ndarray, y: np.ndarray) -> float:
    """Angle between two column vectors.

    Parameters
    ----------
    x: np.ndarray
        Column vector represented as NumPy (3,1) array.
    y: np.ndarray
        Column vector represented as NumPy (3,1) array.

    Returns
    -------
    float
        Angle between the vectors.
    """
    costheta = dot3(x * (1 / norm(x)), y * (1 / norm(y)))
    return acos(bound(costheta))


## Math


def bound(x: float) -> float:
    """Check the input value is in [-1, 1] range.

    Rounds input value to +/-1 if |x| - 1 < SMALL.

    Parameters
    ----------
    x: float
        Input value to be checked

    Returns
    -------
    float
        Value in [-1, 1] range.

    Raises
    ------
    AssertionError
        Input value outside [-1, 1] range.
    """
    if abs(x) > (1 + SMALL):
        raise AssertionError(
            "The value (%f) was unexpectedly too far outside -1 or 1 to "
            "safely bound. Please report this." % x
        )
    if x > 1:
        return 1
    if x < -1:
        return -1
    return x


def radians_equivalent(first: float, second: float, tolerance: float = SMALL) -> bool:
    """Check for angle equivalence.

    Parameters
    ----------
    first: float
        First angle value.
    second:float
        Second angle value.
    tolerance: float, default = SMALL
        Absolute tolerance for the angle difference.

    Returns
    -------
    bool
        True is angles are equivalent.

    """
    diff = sin((first - second) / 2.0)
    return is_small(diff, tolerance)


def isnum(o: Any) -> bool:
    """Check if the input object type is either int or float.

    Parameters
    ----------
    o: Any
        Input object to be checked.

    Returns
    -------
    bool
        If object type is either int or float.
    """
    return isinstance(o, (int, float))


def allnum(lst: Sequence[Any]) -> bool:
    """Check if all object types in the input sequence are either int or float.

    Parameters
    ----------
    o: Sequence[Any]
        Input object sequence to be checked.

    Returns
    -------
    bool
        If all object types in th sequence are either int or float.
    """
    return not [o for o in lst if not isnum(o)]


def is_small(x, tolerance=SMALL) -> bool:
    """Check if input value is 0 within tolerance.

    Parameters
    ----------
    x: float
        Input value to be checked.
    tolerance: float, default = SMALL
        Absolute tolerance.

    Returns
    -------
    bool
        True is the value is 0 within tolerance.
    """
    return isclose(x, 0, abs_tol=tolerance)


def sign(x: float, tolerance: float = SMALL) -> int:
    """Sign function with specified tolerance.

    Parameters
    ----------
    x: float
        Function argument.
    tolerance: float, default = SMALL
        Absolute tolerance.

    Returns
    -------
    int
        1 for positive, -1 for negative values and
        0 if argument equals 0 within tolerance.
    """
    if is_small(x, tolerance):
        return 0
    if x > 0:
        return 1
    # x < 0
    return -1


def normalised(vector: np.ndarray) -> np.ndarray:
    """Normalise vector array.

    Return normalised vector coordinates.

    Parameters
    ----------
    vector : ndarray
        The vector to be normalised.

    Returns
    -------
    ndarray
        Normalised vector.
    """
    return vector * (1.0 / norm(vector))


def zero_round(num):
    """Round to zero if small.

    This is useful to get rid of erroneous minus signs
    resulting from float representation close to zero.

    Parameters
    ----------
    num : number
        The value to be checked for rounding.

    Returns
    -------
    number
        The rounded input value.
    """
    if abs(num) < SMALL:
        num = 0
    return num



"""Module handling constraint information for diffractometer calculations."""
import dataclasses
from enum import Enum
from itertools import zip_longest
from math import degrees, radians
from typing import Callable, Collection, Dict, List, Optional, Tuple, Union

from diffcalc.util import DiffcalcException

_con_category = Enum("_con_category", "DETECTOR REFERENCE SAMPLE")
_con_type = Enum("_con_type", "VALUE VOID")


@dataclasses.dataclass(eq=False)
class _Constraint:
    name: str
    _category: _con_category
    _type: _con_type
    value: Optional[float] = None

    @property
    def active(self) -> bool:
        return self.value is not False and self.value is not None


class Constraints:
    """Collection of angle constraints for diffractometer calculations.

    Three constraints are required for calculations of miller indices and
    the corresponding diffractometer positions. Allowed configurations include
    at most one of the reference and the detector type constraint and up to
    three of the sample type constraints.

    List of the available constraint combinations:

        1 x samp, 1 x ref and 1 x det:  all

        2 x samp and 1 x ref:  chi & phi
                               chi & eta
                               chi & mu
                               mu & eta
                               mu & phi
                               eta & phi

        2 x samp and 1 x det:  chi & phi
                               mu & eta
                               mu & phi
                               mu & chi
                               eta & phi
                               eta & chi
                               bisect & mu
                               bisect & eta
                               bisect & omega

        3 x samp:              eta, chi & phi
                               mu, chi & phi
                               mu, eta & phi
                               mu, eta & chi
    """

    def __init__(
        self,
        constraints: Collection[Union[Tuple[str, float], str]] = None,
        indegrees: bool = True,
    ):
        """Object for setting diffractometer angle constraints."""
        self._delta = _Constraint("delta", _con_category.DETECTOR, _con_type.VALUE)
        self._nu = _Constraint("nu", _con_category.DETECTOR, _con_type.VALUE)
        self._qaz = _Constraint("qaz", _con_category.DETECTOR, _con_type.VALUE)
        self._naz = _Constraint("naz", _con_category.DETECTOR, _con_type.VALUE)

        self._a_eq_b = _Constraint("a_eq_b", _con_category.REFERENCE, _con_type.VOID)
        self._alpha = _Constraint("alpha", _con_category.REFERENCE, _con_type.VALUE)
        self._beta = _Constraint("beta", _con_category.REFERENCE, _con_type.VALUE)
        self._psi = _Constraint("psi", _con_category.REFERENCE, _con_type.VALUE)
        self._bin_eq_bout = _Constraint(
            "bin_eq_bout", _con_category.REFERENCE, _con_type.VOID
        )
        self._betain = _Constraint("betain", _con_category.REFERENCE, _con_type.VALUE)
        self._betaout = _Constraint("betaout", _con_category.REFERENCE, _con_type.VALUE)

        self._mu = _Constraint("mu", _con_category.SAMPLE, _con_type.VALUE)
        self._eta = _Constraint("eta", _con_category.SAMPLE, _con_type.VALUE)
        self._chi = _Constraint("chi", _con_category.SAMPLE, _con_type.VALUE)
        self._phi = _Constraint("phi", _con_category.SAMPLE, _con_type.VALUE)
        self._bisect = _Constraint("bisect", _con_category.SAMPLE, _con_type.VOID)
        self._omega = _Constraint("omega", _con_category.SAMPLE, _con_type.VALUE)

        self._all: Tuple[_Constraint, ...] = (
            self._delta,
            self._nu,
            self._qaz,
            self._naz,
            self._a_eq_b,
            self._alpha,
            self._beta,
            self._psi,
            self._bin_eq_bout,
            self._betain,
            self._betaout,
            self._mu,
            self._eta,
            self._chi,
            self._phi,
            self._bisect,
            self._omega,
        )
        self.indegrees = indegrees
        if constraints is not None:
            if isinstance(constraints, dict):
                self.asdict = constraints
            elif isinstance(constraints, tuple):
                self.astuple = constraints
            elif isinstance(constraints, (list, set)):
                self.astuple = tuple(constraints)
            else:
                raise DiffcalcException(
                    f"Invalid constraint parameter type: {type(constraints)}"
                )

    def __str__(self) -> str:
        """Output text representation of active constraint set.

        Returns
        -------
        str
            Table representation of the available constraints with a list of
            constrained angle names and values.
        """
        lines = []
        lines.extend(self._build_display_table_lines())
        lines.append("")
        lines.extend(self._report_constraints_lines())
        lines.append("")
        if self.is_fully_constrained() and not self.is_current_mode_implemented():
            lines.append("    Sorry, this constraint combination is not implemented.")
        return "\n".join(lines)

    @property
    def _constrained(self):
        return tuple(con for con in self._all if con.active)

    @property
    def _detector(self):
        return {
            con.name: con.value
            for con in self._all
            if con.active and con._category is _con_category.DETECTOR
        }

    @property
    def _reference(self):
        return {
            con.name: con.value
            for con in self._all
            if con.active and con._category is _con_category.REFERENCE
        }

    @property
    def _sample(self):
        return {
            con.name: con.value
            for con in self._all
            if con.active and con._category is _con_category.SAMPLE
        }

    @property
    def all(self) -> Dict[str, Union[float, bool, None]]:
        """Get all angle names and values.

        Returns
        -------
        Dict[str, Union[float, bool, None]]
            Dictionary with all angle names and values.
        """
        return {con.name: getattr(self, con.name) for con in self._all}

    @property
    def asdict(self) -> Dict[str, Union[float, bool]]:
        """Get all constrained angle names and values.

        Returns
        -------
        Dict[str, Union[float, bool]]
            Dictionary with all constrained angle names and values.
        """
        return {con.name: getattr(self, con.name) for con in self._all if con.active}

    @asdict.setter
    def asdict(self, constraints):
        assert isinstance(constraints, dict)
        self.clear()
        if constraints is None:
            return
        for con_name, con_value in constraints.items():
            if hasattr(self, con_name) and isinstance(
                getattr(self, "_" + con_name), _Constraint
            ):
                setattr(self, con_name, con_value)
            else:
                raise DiffcalcException(f"Invalid constraint name: {con_name}")

    @property
    def astuple(self) -> Tuple[Union[Tuple[str, float], str], ...]:
        """Get all constrained angle names and values.

        Returns
        -------
        Tuple[Union[Tuple[str, float], str], ...]
            Tuple with all constrained angle names and values.
        """
        res = []
        for con in self._constrained:
            if con._type is _con_type.VALUE:
                res.append((con.name, getattr(self, con.name)))
            elif con._type is _con_type.VOID:
                res.append(con.name)
            else:
                raise DiffcalcException(
                    f"Invalid {con.name} constraint type found: {type(con._type)}"
                )
        return tuple(res)

    @astuple.setter
    def astuple(self, constraints: Tuple[Union[Tuple[str, float], str], ...]) -> None:
        assert isinstance(constraints, tuple)
        self.clear()
        if constraints is None:
            return
        for el in constraints:
            if (
                isinstance(el, str)
                and hasattr(self, el)
                and isinstance(getattr(self, "_" + el), _Constraint)
            ):
                setattr(self, el, True)
            elif (
                isinstance(el, (type(("a", 1)), type(("a", 1.0))))
                and hasattr(self, el[0])
                and isinstance(getattr(self, "_" + el[0]), _Constraint)
            ):
                setattr(self, *el)
            else:
                raise DiffcalcException(f"Invalid constraint parameter: {el}")

    def _get_factory(self, con: _Constraint) -> Callable[[], Union[float, bool, None]]:
        def _get_constraint() -> Optional[Union[float, bool, None]]:
            if con.value is None:
                return None
            if isinstance(con.value, bool):
                return con.value
            elif isinstance(con.value, (int, float)):
                return degrees(con.value) if self.indegrees else con.value
            else:
                raise DiffcalcException(
                    f"Invalid {con.name} value type: {type(con.value)}"
                )

        return _get_constraint

    def _set_factory(
        self, con: _Constraint
    ) -> Callable[[Union[float, bool, None]], None]:
        def _set_value(val: Union[float, bool]) -> None:
            if isinstance(val, bool):
                if con._type is _con_type.VOID:
                    con.value = val
                    return
                else:
                    raise DiffcalcException(
                        f"Constraint {con.name} requires numerical value. "
                        f"Found {type(val)} instead."
                    )
            if con._type is _con_type.VALUE:
                try:
                    con.value = radians(float(val)) if self.indegrees else float(val)
                    return
                except ValueError:
                    raise DiffcalcException(
                        f"Constraint {con.name} requires numerical value. "
                        f"Found {type(val)} instead."
                    )
            raise DiffcalcException(
                f"Constraint {con.name} requires boolean value. "
                f"Found {type(val)} instead."
            )

        def _set_constraint(val: Union[float, bool, None]) -> None:
            if val is None or val is False:
                con.value = None
                return
            active_con = {c for c in self._constrained if c._category is con._category}
            num_active_con = len(active_con)
            if con in active_con:
                _set_value(val)
                return
            # Check if there's free constraint slot in a given constraint category and overall
            if not self.is_fully_constrained(con) and not self.is_fully_constrained():
                _set_value(val)
            # We don't have empty slot for a new constraints.
            # We need to replace on of the active constraints
            elif num_active_con > 1:
                # If there's already more than one constraint set for the given category.
                # We don't know which one we should replace.
                raise DiffcalcException(
                    f"Cannot set {con.name} constraint. First un-constrain one of the\n"
                    f"angles {', '.join(sorted(c.name for c in self._constrained))}."
                )
            elif num_active_con == 0:
                # We need to replace a constraint from other category.
                # We don't know which one to replace
                raise DiffcalcException(
                    f"Cannot set {con.name} constraint. First un-constrain one of the\n"
                    f"angles {', '.join(sorted(c.name for c in self._constrained))}."
                )
            elif num_active_con == 1:
                # If we have only one constraint in the requested category.
                # We'll replace it with the new constraint.
                existing_con = active_con.pop()
                existing_con.value = None
            _set_value(val)

        return _set_constraint

    def _del_factory(self, con: _Constraint) -> Callable[[], None]:
        def _del_constraint() -> None:
            con.value = None

        return _del_constraint

    @property
    def delta(self) -> Union[float, None]:
        """Constraint for delta angle."""
        return self._get_factory(self._delta)()

    @delta.setter
    def delta(self, val):
        return self._set_factory(self._delta)(val)

    @delta.deleter
    def delta(self):
        return self._del_factory(self._delta)()

    @property
    def nu(self):
        """Constraint for nu angle."""
        return self._get_factory(self._nu)()

    @nu.setter
    def nu(self, val):
        return self._set_factory(self._nu)(val)

    @nu.deleter
    def nu(self):
        return self._del_factory(self._nu)()

    @property
    def qaz(self):
        """Constraint for qaz angle."""
        return self._get_factory(self._qaz)()

    @qaz.setter
    def qaz(self, val):
        return self._set_factory(self._qaz)(val)

    @qaz.deleter
    def qaz(self):
        return self._del_factory(self._qaz)()

    @property
    def naz(self):
        """Constraint for naz angle."""
        return self._get_factory(self._naz)()

    @naz.setter
    def naz(self, val):
        return self._set_factory(self._naz)(val)

    @naz.deleter
    def naz(self):
        return self._del_factory(self._naz)()

    @property
    def a_eq_b(self):
        """Constraint for setting alpha = beta."""
        return self._get_factory(self._a_eq_b)()

    @a_eq_b.setter
    def a_eq_b(self, val):
        return self._set_factory(self._a_eq_b)(val)

    @a_eq_b.deleter
    def a_eq_b(self):
        return self._del_factory(self._a_eq_b)()

    @property
    def alpha(self):
        """Constraint for alpha angle."""
        return self._get_factory(self._alpha)()

    @alpha.setter
    def alpha(self, val):
        return self._set_factory(self._alpha)(val)

    @alpha.deleter
    def alpha(self):
        return self._del_factory(self._alpha)()

    @property
    def beta(self):
        """Constraint for beta angle."""
        return self._get_factory(self._beta)()

    @beta.setter
    def beta(self, val):
        return self._set_factory(self._beta)(val)

    @beta.deleter
    def beta(self):
        return self._del_factory(self._beta)()

    @property
    def psi(self):
        """Constraint for psi angle."""
        return self._get_factory(self._psi)()

    @psi.setter
    def psi(self, val):
        return self._set_factory(self._psi)(val)

    @psi.deleter
    def psi(self):
        return self._del_factory(self._psi)()

    @property
    def bin_eq_bout(self):
        """Constraint for betain = betaout."""
        return self._get_factory(self._bin_eq_bout)()

    @bin_eq_bout.setter
    def bin_eq_bout(self, val):
        return self._set_factory(self._bin_eq_bout)(val)

    @bin_eq_bout.deleter
    def bin_eq_bout(self):
        return self._del_factory(self._bin_eq_bout)()

    @property
    def betain(self):
        """Constraint for betain angle."""
        return self._get_factory(self._betain)()

    @betain.setter
    def betain(self, val):
        return self._set_factory(self._betain)(val)

    @betain.deleter
    def betain(self):
        return self._del_factory(self._betain)()

    @property
    def betaout(self):
        """Constraint for betaout angle."""
        return self._get_factory(self._betaout)()

    @betaout.setter
    def betaout(self, val):
        return self._set_factory(self._betaout)(val)

    @betaout.deleter
    def betaout(self):
        return self._del_factory(self._betaout)()

    @property
    def mu(self):
        """Constraint for mu angle."""
        return self._get_factory(self._mu)()

    @mu.setter
    def mu(self, val):
        return self._set_factory(self._mu)(val)

    @mu.deleter
    def mu(self):
        return self._del_factory(self._mu)()

    @property
    def eta(self):
        """Constraint for eta angle."""
        return self._get_factory(self._eta)()

    @eta.setter
    def eta(self, val):
        return self._set_factory(self._eta)(val)

    @eta.deleter
    def eta(self):
        return self._del_factory(self._eta)()

    @property
    def chi(self):
        """Constraint for chi angle."""
        return self._get_factory(self._chi)()

    @chi.setter
    def chi(self, val):
        return self._set_factory(self._chi)(val)

    @chi.deleter
    def chi(self):
        return self._del_factory(self._chi)()

    @property
    def phi(self):
        """Constraint for phi angle."""
        return self._get_factory(self._phi)()

    @phi.setter
    def phi(self, val):
        return self._set_factory(self._phi)(val)

    @phi.deleter
    def phi(self):
        return self._del_factory(self._phi)()

    @property
    def bisect(self):
        """Constraint for bisect mode."""
        return self._get_factory(self._bisect)()

    @bisect.setter
    def bisect(self, val):
        return self._set_factory(self._bisect)(val)

    @bisect.deleter
    def bisect(self):
        return self._del_factory(self._bisect)()

    @property
    def omega(self):
        """Constraint for omega angle."""
        return self._get_factory(self._omega)()

    @omega.setter
    def omega(self, val):
        return self._set_factory(self._omega)(val)

    @omega.deleter
    def omega(self):
        return self._del_factory(self._omega)()

    def _build_display_table_lines(self) -> List[str]:
        constraint_types = [
            (self._delta, self._nu, self._qaz, self._naz),
            (
                self._a_eq_b,
                self._alpha,
                self._beta,
                self._psi,
                self._bin_eq_bout,
                self._betain,
                self._betaout,
            ),
            (self._mu, self._eta, self._chi, self._phi, self._bisect, self._omega),
        ]
        max_name_width = max([len(con.name) for con in self._all])

        cells = []

        header_cells = []
        header_cells.append("    " + "DET".ljust(max_name_width))
        header_cells.append("    " + "REF".ljust(max_name_width))
        header_cells.append("    " + "SAMP")
        cells.append(header_cells)

        underline_cells = ["    " + "-" * max_name_width] * len(constraint_types)
        cells.append(underline_cells)

        for con_line in zip_longest(*constraint_types):
            row_cells = []
            for con in con_line:
                name = con.name if con is not None else ""
                row_cells.append("   " if con is None or not con.active else "-->")
                row_cells.append(("%-" + str(max_name_width) + "s") % name)
            cells.append(row_cells)

        lines = [" ".join(row_cells).rstrip() for row_cells in cells]
        return lines

    def _report_constraint(self, con: _Constraint) -> str:
        val = getattr(self, con.name)
        if con._type is _con_type.VOID:
            return "    %s" % con.name
        else:
            return f"    {con.name:<5}: {val:.4f}"

    def _report_constraints_lines(self) -> List[str]:
        lines = []
        required = 3 - len(self._constrained)
        if required == 0:
            pass
        elif required == 1:
            lines.append("!   1 more constraint required")
        else:
            lines.append("!   %d more constraints required" % required)
        lines.extend([self._report_constraint(con) for con in self._all if con.active])
        return lines

    @classmethod
    def asdegrees(cls, constraints: "Constraints") -> "Constraints":
        """Create new Constraints object with angles in degrees.

        Parameters
        ----------
        constraints: Constraints
            Input Constraints object

        Returns
        -------
        Constraints
            New Constraints object with angles in degrees.
        """
        res = cls(constraints.asdict, indegrees=constraints.indegrees)
        res.indegrees = True
        return res

    @classmethod
    def asradians(cls, constraints: "Constraints") -> "Constraints":
        """Create new Constraints object with angles in radians.

        Parameters
        ----------
        constraints: Constraints
            Input Position object

        Returns
        -------
        Constraints
            New Constraints object with angles in radians.
        """
        res = cls(constraints.asdict, indegrees=constraints.indegrees)
        res.indegrees = False
        return res

    def is_fully_constrained(self, con: Optional[_Constraint] = None) -> bool:
        """Check if configuration is fully constrained.

        Parameters
        ----------
        con: _Constraint, default = None
            Check if there are available constraints is the same category as the
            input constraint. If parameter is None, check for all constraint
            categories.

        Returns
        -------
        bool:
            True if there aren't any constraints available either in the input
            constraint category or no constraints are available.
        """
        if con is None:
            return len(self._constrained) >= 3

        _max_constrained = {
            _con_category.DETECTOR: 1,
            _con_category.REFERENCE: 1,
            _con_category.SAMPLE: 3,
        }
        count_constrained = len(
            {c for c in self._constrained if c._category is con._category}
        )
        return count_constrained >= _max_constrained[con._category]

    def is_current_mode_implemented(self) -> bool:
        """Check if current constraint set is implemented.

        Configuration needs to be fully constraint for this method to work.

        Returns
        -------
        bool:
            True if current constraint set is supported.
        """
        if not self.is_fully_constrained():
            raise ValueError("Three constraints required")

        count_detector = len(
            {c for c in self._constrained if c._category is _con_category.DETECTOR}
        )
        count_reference = len(
            {c for c in self._constrained if c._category is _con_category.REFERENCE}
        )
        count_sample = len(
            {c for c in self._constrained if c._category is _con_category.SAMPLE}
        )
        if count_sample == 3:
            if (
                set(self._constrained) == {self._chi, self._phi, self._eta}
                or set(self._constrained) == {self._chi, self._phi, self._mu}
                or set(self._constrained) == {self._chi, self._eta, self._mu}
                or set(self._constrained) == {self._phi, self._eta, self._mu}
            ):
                return True
            return False

        if count_sample == 1:
            return self._omega not in set(
                self._constrained
            ) and self._bisect not in set(self._constrained)

        if count_reference == 1:
            return (
                {self._chi, self._phi}.issubset(self._constrained)
                or {self._chi, self._eta}.issubset(self._constrained)
                or {self._chi, self._mu}.issubset(self._constrained)
                or {self._mu, self._eta}.issubset(self._constrained)
                or {self._mu, self._phi}.issubset(self._constrained)
                or {self._eta, self._phi}.issubset(self._constrained)
            )

        if count_detector == 1:
            return (
                {self._chi, self._phi}.issubset(self._constrained)
                or {self._mu, self._eta}.issubset(self._constrained)
                or {self._mu, self._phi}.issubset(self._constrained)
                or {self._mu, self._chi}.issubset(self._constrained)
                or {self._eta, self._phi}.issubset(self._constrained)
                or {self._eta, self._chi}.issubset(self._constrained)
                or {self._mu, self._bisect}.issubset(self._constrained)
                or {self._eta, self._bisect}.issubset(self._constrained)
                or {self._omega, self._bisect}.issubset(self._constrained)
            )

        return False

    def clear(self) -> None:
        """Remove all constraints."""
        for con in self._all:
            delattr(self, con.name)



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

from math import pi

from diffcalc.ub.calc import UBCalculation
from numpy import array

from tests.diffcalc.scenarios import PosFromI16sEuler
from tests.tools import matrixeq_

# newub 'cubic'                   <-->  reffile('cubic)
# setlat 'cubic' 1 1 1 90 90 90   <-->  latt([1,1,1,90,90,90])
# pos wl 1                        <-->  BLi.setWavelength(1)
#                                <-->  c2th([0,0,1]) --> 60
# pos sixc [0 60 0 30 1 1]        <-->  pos euler [1 1 30 0 60 0]
# addref 1 0 0                    <-->  saveref('100',[1, 0, 0])
# pos chi 91                      <-->
# addref 0 0 1                    <-->  saveref('100',[1, 0, 0]) ; showref()
#                                      ubm('100','001')
#                                      ubm() ->
# array('d', [0.9996954135095477, -0.01745240643728364, -0.017449748351250637,
# 0.01744974835125045, 0.9998476951563913, -0.0003045864904520898,
# 0.017452406437283505, -1.1135499981271473e-16, 0.9998476951563912])


UB1 = (
    array(
        (
            (0.9996954135095477, -0.01745240643728364, -0.017449748351250637),
            (0.01744974835125045, 0.9998476951563913, -0.0003045864904520898),
            (0.017452406437283505, -1.1135499981271473e-16, 0.9998476951563912),
        )
    )
    * (2 * pi)
)

EN1 = 12.39842
REF1a = PosFromI16sEuler(1, 1, 30, 0, 60, 0)
REF1b = PosFromI16sEuler(1, 91, 30, 0, 60, 0)


def testAgainstI16Results():
    ubcalc = UBCalculation("cubcalc")
    ubcalc.set_lattice("latt", 1, 1, 1, 90, 90, 90)
    ubcalc.add_reflection((1, 0, 0), REF1a, EN1, "100")
    ubcalc.add_reflection((0, 0, 1), REF1b, EN1, "001")
    ubcalc.calc_ub()
    matrixeq_(ubcalc.UB, UB1)



###
# Copyright 2008-2011 Diamond Light Source Ltd.
# This file is part of Diffcalc.
#
# Diffcalc is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Diffcalc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Diffcalc.  If not, see <http://www.gnu.org/licenses/>.
###

# TODO: class largely copied from test_calc

from collections import namedtuple

import pytest
from diffcalc.hkl.geometry import Position
from diffcalc.ub.calc import UBCalculation
from diffcalc.ub.crystal import Crystal
from diffcalc.util import DiffcalcException, I
from numpy import array

from tests.diffcalc.hkl.test_calc import _BaseTest
from tests.tools import matrixeq_


class TestSurfaceNormalVerticalCubic(_BaseTest):
    def setup_method(self):

        _BaseTest.setup_method(self)
        self.constraints.asdict = {"a_eq_b": True, "mu": -90, "eta": 0}
        self.wavelength = 1

    def _configure_ub(self):
        self.ubcalc.n_phi = (1, 0, 0)
        self.ubcalc.set_lattice("xtal", 1)
        self.ubcalc.set_u(I)

    def _check(self, hkl, pos, virtual_expected={}):
        if pos is not None:
            self._check_angles_to_hkl(
                "", 999, 999, hkl, pos, self.wavelength, virtual_expected
            )
        self._check_hkl_to_angles(
            "", 999, 999, hkl, pos, self.wavelength, virtual_expected
        )

    def testHkl001(self):
        pos = Position(
            mu=-90,
            delta=60,
            nu=0,
            eta=0,
            chi=90 + 30,
            phi=90,
        )
        self._check((0, 0, 1), pos, {"betain": 30, "betaout": 30})

    def testHkl011(self):
        pos = Position(
            mu=-90,
            delta=90,
            nu=0,
            eta=0,
            chi=90,
            phi=90,
        )
        self._check((0, 1, 1), pos, {"betain": 0, "betaout": 90})

    def testHkl010(self):
        pos = Position(
            mu=-90,
            delta=60,
            nu=0,
            eta=0,
            chi=30,
            phi=90,
        )
        self._check((0, 1, 0), pos, {"betain": -60, "betaout": 60})

    @pytest.mark.xfail(raises=DiffcalcException)
    def testHkl100(self):
        self._check((1, 0, 0), None, {"alpha": 30, "beta": 30})

    def testHkl110(self):
        pos = Position(
            mu=-90,
            delta=90,
            nu=0,
            eta=0,
            chi=45,
            phi=45,
        )
        self._check((1, 1, 0), pos, {"alpha": 30, "beta": 30})


# Primary and secondary reflections found with the help of DDIF on Diamond's
# i07 on Jan 27 2010
WillPos = namedtuple("WillPos", ["delta", "gamma", "omegah", "phi"])

HKL0 = 2, 19, 32
REF0 = WillPos(delta=21.975, gamma=4.419, omegah=2, phi=326.2)

HKL1 = 0, 7, 22
REF1 = WillPos(delta=11.292, gamma=2.844, omegah=2, phi=124.1)

WAVELENGTH = 0.6358
ENERGY = 12.39842 / WAVELENGTH


# This is the version that Diffcalc comes up with ( see following test)
U_DIFFCALC = array(
    [
        [-0.7178876, 0.6643924, -0.2078944],
        [-0.6559596, -0.5455572, 0.5216170],
        [0.2331402, 0.5108327, 0.8274634],
    ]
)


# class WillmottHorizontalGeometry(VliegGeometry):
#
#    def __init__(self):
#        VliegGeometry.__init__(self,
#                    name='willmott_horizontal',
#                    supported_mode_groups=[],
#                    fixed_parameters={},
#                    gamma_location='base'
#                    )
#
#    def physical_angles_to_internal_position(self, physicalAngles):
#        assert (len(physicalAngles) == 4), "Wrong length of input list"
#        return WillPos(*physicalAngles)
#
#    def internal_position_to_physical_angles(self, internalPosition):
#        return internalPosition.totuple()


def willmott_to_you_fixed_mu_eta(pos):
    pos = Position(
        mu=-90,
        delta=pos.delta,
        nu=pos.gamma,
        eta=0,
        chi=90 + pos.omegah,
        phi=-90 - pos.phi,
    )
    return pos


class TestUBCalculationWithWillmotStrategy_Si_5_5_12_FixedMuEta:
    def testAgainstResultsFromJan_27_2010(self):
        self.ubcalc = UBCalculation("test")
        self.ubcalc.set_lattice("Si_5_5_12", 7.68, 53.48, 75.63, 90, 90, 90)
        self.ubcalc.add_reflection(
            HKL0,
            willmott_to_you_fixed_mu_eta(REF0),
            ENERGY,
            "ref0",
        )
        self.ubcalc.add_reflection(
            HKL1,
            willmott_to_you_fixed_mu_eta(REF1),
            ENERGY,
            "ref1",
        )
        self.ubcalc.calc_ub()
        print("U: ", self.ubcalc.U)
        print("UB: ", self.ubcalc.UB)
        matrixeq_(self.ubcalc.U, U_DIFFCALC)


class TestFixedMuEta(_BaseTest):
    def setup_method(self):
        _BaseTest.setup_method(self)
        self._configure_constraints()
        self.wavelength = 0.6358

    def _configure_constraints(self):
        self.constraints.asdict = {"alpha": 2, "mu": -90, "eta": 0}

    def _convert_willmott_pos(self, willmott_pos):
        return willmott_to_you_fixed_mu_eta(willmott_pos)

    def _configure_ub(self):
        self.ubcalc.set_lattice("xtal", 7.68, 53.48, 75.63, 90, 90, 90)
        self.ubcalc.set_u(U_DIFFCALC)

    def _check(self, hkl, pos, virtual_expected={}):
        self._check_angles_to_hkl(
            "", 999, 999, hkl, pos, self.wavelength, virtual_expected
        )
        self._check_hkl_to_angles(
            "", 999, 999, hkl, pos, self.wavelength, virtual_expected
        )

    def testHkl_2_19_32_found_orientation_setting(self):
        """Check that the or0 reflection maps back to the assumed hkl"""
        self.places = 2
        self._check_angles_to_hkl(
            "",
            999,
            999,
            HKL0,
            self._convert_willmott_pos(REF0),
            self.wavelength,
            {"alpha": 2},
        )

    def testHkl_0_7_22_found_orientation_setting(self):
        """Check that the or1 reflection maps back to the assumed hkl"""
        self.places = 0
        self._check_angles_to_hkl(
            "",
            999,
            999,
            HKL1,
            self._convert_willmott_pos(REF1),
            self.wavelength,
            {"alpha": 2},
        )

    def testHkl_2_19_32_calculated_from_DDIF(self):
        self.places = 3
        willpos = WillPos(delta=21.974, gamma=4.419, omegah=2, phi=-33.803)
        self._check((2, 19, 32), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_0_7_22_calculated_from_DDIF(self):
        self.places = 3
        willpos = WillPos(
            delta=11.241801854649, gamma=-3.038407637123, omegah=2, phi=-86.56344250267
        )
        self._check((0, 7, 22), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_2_m5_12_calculated_from_DDIF(self):
        self.places = 3
        willpos = WillPos(delta=5.224, gamma=10.415, omegah=2, phi=-1.972)
        self._check((2, -5, 12), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_2_19_32_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(
            delta=21.974032376045, gamma=4.418955754003, omegah=2, phi=-33.80254
        )
        self._check((2, 19, 32), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_0_7_22_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(
            delta=11.241801854649, gamma=-3.038407637123, omegah=2, phi=-86.563442502670
        )
        self._check((0, 7, 22), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_2_m5_12_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(
            delta=5.223972025344, gamma=10.415435905622, omegah=2, phi=-90 + 88.02751
        )
        self._check((2, -5, 12), self._convert_willmott_pos(willpos), {"alpha": 2})


###############################################################################


def willmott_to_you_fixed_mu_chi(pos):
    return Position(
        mu=-0,
        delta=pos.delta,
        nu=pos.gamma,
        eta=pos.omegah,
        chi=90,
        phi=-pos.phi,
    )


class TestUBCalculationWithWillmotStrategy_Si_5_5_12_FixedMuChi:
    def testAgainstResultsFromJan_27_2010(self):
        self.ubcalc = UBCalculation("test")
        self.ubcalc.set_lattice("Si_5_5_12", 7.68, 53.48, 75.63, 90, 90, 90)
        self.ubcalc.add_reflection(
            HKL0,
            willmott_to_you_fixed_mu_chi(REF0),
            ENERGY,
            "ref0",
        )
        self.ubcalc.add_reflection(
            HKL1,
            willmott_to_you_fixed_mu_chi(REF1),
            ENERGY,
            "ref1",
        )
        self.ubcalc.calc_ub()
        print("U: ", self.ubcalc.U)
        print("UB: ", self.ubcalc.UB)
        matrixeq_(self.ubcalc.U, U_DIFFCALC)


class Test_Fixed_Mu_Chi(TestFixedMuEta):
    def _configure_constraints(self):
        self.constraints.asdict = {"alpha": 2, "mu": 0, "chi": 90}

    def _convert_willmott_pos(self, willmott_pos):
        return willmott_to_you_fixed_mu_chi(willmott_pos)


def willmott_to_you_fixed_eta_chi(pos):
    return Position(
        mu=pos.omegah,
        delta=-pos.gamma,
        nu=pos.delta,
        eta=0,
        chi=0,
        phi=-pos.phi,
    )


class Test_Fixed_Eta_Chi(TestFixedMuEta):
    def _configure_constraints(self):
        self.constraints.asdict = {"alpha": 2, "eta": 0, "chi": 0}

    def _convert_willmott_pos(self, willmott_pos):
        return willmott_to_you_fixed_eta_chi(willmott_pos)

    def testHkl_2_19_32_found_orientation_setting(self):
        pytest.skip()

    def testHkl_0_7_22_found_orientation_setting(self):
        pytest.skip()

    def testHkl_2_19_32_calculated_from_DDIF(self):
        pytest.skip()

    def testHkl_0_7_22_calculated_from_DDIF(self):
        pytest.skip()

    def testHkl_2_m5_12_calculated_from_DDIF(self):
        pytest.skip()

    def testHkl_2_19_32_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(delta=22.0332862, gamma=-4.0973643, omegah=2, phi=64.0273584)
        self._check((2, 19, 32), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_0_7_22_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(delta=11.2572236, gamma=-2.9800571, omegah=2, phi=-86.5634425)
        self._check((0, 7, 22), self._convert_willmott_pos(willpos), {"alpha": 2})

    def testHkl_2_m5_12_calculated_predicted_with_diffcalc_and_found(self):
        willpos = WillPos(delta=5.3109941, gamma=-10.3716944, omegah=2, phi=167.0041454)
        self._check((2, -5, 12), self._convert_willmott_pos(willpos), {"alpha": 2})


# Primary and secondary reflections found with the help of DDIF on Diamond's
# i07 on Jan 28/29 2010


Pt531_HKL0 = -1.000, 1.000, 6.0000
Pt531_REF0 = WillPos(delta=9.3971025, gamma=16.1812303, omegah=2, phi=-52.1392905)

Pt531_HKL1 = -2.000, -1.000, 7.0000
Pt531_REF1 = WillPos(delta=11.0126958, gamma=-11.8636128, omegah=2, phi=40.3803393)
Pt531_REF12 = WillPos(delta=11.0126958, gamma=11.8636128, omegah=2, phi=-121.2155975)
Pt531_HKL2 = 1, 1, 9
Pt531_REF2 = WillPos(delta=14.1881617, gamma=7.7585939, omegah=2, phi=23.0203132)
Pt531_REF22 = WillPos(delta=14.1881617, gamma=-7.7585939, omegah=2, phi=-183.465146)
Pt531_WAVELENGTH = 0.6358

# This is U matrix displayed by DDIF
U_FROM_DDIF = array(
    [
        [-0.00312594, -0.00063417, 0.99999491],
        [0.99999229, -0.00237817, 0.00312443],
        [0.00237618, 0.99999697, 0.00064159],
    ]
)

# This is the version that Diffcalc comes up with ( see following test)
Pt531_U_DIFFCALC = array(
    [
        [-0.0023763, -0.9999970, -0.0006416],
        [0.9999923, -0.0023783, 0.0031244],
        [-0.0031259, -0.0006342, 0.9999949],
    ]
)


class TestUBCalculation_Pt531_FixedMuChi:
    def testAgainstResultsFromJan_28_2010(self):
        self.ubcalc = UBCalculation("test")
        self.ubcalc.set_lattice("Pt531", 6.204, 4.806, 23.215, 90, 90, 49.8)

        self.ubcalc.add_reflection(
            Pt531_HKL0,
            willmott_to_you_fixed_mu_chi(Pt531_REF0),
            12.39842 / Pt531_WAVELENGTH,
            "ref0",
        )
        self.ubcalc.add_reflection(
            Pt531_HKL1,
            willmott_to_you_fixed_mu_chi(Pt531_REF1),
            12.39842 / Pt531_WAVELENGTH,
            "ref1",
        )
        self.ubcalc.calc_ub()
        print("U: ", self.ubcalc.U)
        print("UB: ", self.ubcalc.UB)
        matrixeq_(self.ubcalc.U, Pt531_U_DIFFCALC)


class Test_Pt531_FixedMuChi(_BaseTest):
    def setup_method(self):
        _BaseTest.setup_method(self)
        self._configure_constraints()
        self.wavelength = Pt531_WAVELENGTH
        CUT = Crystal("Pt531", 6.204, 4.806, 23.215, 90, 90, 49.8)
        B = CUT.B
        self.UB = Pt531_U_DIFFCALC @ B

    def _configure_constraints(self):
        self.constraints.asdict = {"alpha": 2, "mu": 0, "chi": 90}

    def _convert_willmott_pos(self, willmott_pos):
        return willmott_to_you_fixed_mu_chi(willmott_pos)

    def _configure_ub(self):
        self.ubcalc.set_lattice("Pt531", 6.204, 4.806, 23.215, 90, 90, 49.8)
        self.ubcalc.set_u(Pt531_U_DIFFCALC)

    def _check(self, hkl, pos, virtual_expected={}, fails=False):
        self._check_angles_to_hkl(
            "", 999, 999, hkl, pos, self.wavelength, virtual_expected
        )
        if fails:
            self._check_hkl_to_angles_fails(
                "", 999, 999, hkl, pos, self.wavelength, virtual_expected
            )
        else:
            self._check_hkl_to_angles(
                "", 999, 999, hkl, pos, self.wavelength, virtual_expected
            )

    def testHkl_0_found_orientation_setting(self):
        """Check that the or0 reflection maps back to the assumed hkl"""
        self.places = 1
        self._check_angles_to_hkl(
            "",
            999,
            999,
            Pt531_HKL0,
            self._convert_willmott_pos(Pt531_REF0),
            self.wavelength,
            {"alpha": 2},
        )

    def testHkl_1_found_orientation_setting(self):
        """Check that the or1 reflection maps back to the assumed hkl"""
        self.places = 0
        self._check_angles_to_hkl(
            "",
            999,
            999,
            Pt531_HKL1,
            self._convert_willmott_pos(Pt531_REF1),
            self.wavelength,
            {"alpha": 2},
        )

    def testHkl_0_calculated_from_DDIF(self):
        self.places = 7
        pos_expected = self._convert_willmott_pos(Pt531_REF0)
        self._check(Pt531_HKL0, pos_expected, {"alpha": 2})

    def testHkl_1_calculated_from_DDIF(self):
        self.places = 7
        self._check(Pt531_HKL1, self._convert_willmott_pos(Pt531_REF1), {"alpha": 2})

    def testHkl_2_calculated_from_DDIF(self):
        self.places = 5
        self._check(Pt531_HKL2, self._convert_willmott_pos(Pt531_REF2), {"alpha": 2})

    def testHkl_2_m1_0_16(self):
        self.places = 5
        pos = WillPos(delta=25.7990976, gamma=-6.2413545, omegah=2, phi=47.4624380)
        #        pos.phi -= 360
        self._check((-1, 0, 16), self._convert_willmott_pos(pos), {"alpha": 2})


class Test_Pt531_Fixed_Mu_eta_(Test_Pt531_FixedMuChi):
    def _configure_constraints(self):
        self.constraints.asdict = {"alpha": 2, "mu": -90, "eta": 0}

    def _convert_willmott_pos(self, willmott_pos):
        return willmott_to_you_fixed_mu_eta(willmott_pos)

    def testHkl_1_calculated_from_DDIF(self):
        self.places = 7
        self._check(Pt531_HKL1, self._convert_willmott_pos(Pt531_REF12), {"alpha": 2})

    def testHkl_2_calculated_from_DDIF(self):
        self.places = 5
        self._check(Pt531_HKL2, self._convert_willmott_pos(Pt531_REF22), {"alpha": 2})

    def testHkl_2_m1_0_16(self):
        self.places = 5
        pos = WillPos(delta=25.7990976, gamma=6.2413545, omegah=2, phi=-47.4949600)
        #        pos.phi -= 360
        self._check((-1, 0, 16), self._convert_willmott_pos(pos), {"alpha": 2})
